from functools import wraps

import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets

from qtextras import ParameterEditor, RunOptions, bindInteractorOptions as bind


class LAST_RESULT:
    """
    Just for testing purposes
    """

    value = None


def printResult(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        func(*args, **kwargs)
        # False positive: nullptr / None is allowed
        # noinspection PyTypeChecker
        QtWidgets.QMessageBox.information(
            None,
            "Function Run!",
            f"Func result: {LAST_RESULT.value}",
        )

    return wrapper


@printResult
def funcOutsideClassUpdateOnBtn(a=5, b=6):
    """
    Function description. This will appear in the hover menu.
    """
    LAST_RESULT.value = a + b


class MyClass(ParameterEditor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.registerFunction(
            funcOutsideClassUpdateOnBtn, runOptions=RunOptions.ON_BUTTON
        )

        self.registerFunction(self.updateOnChanging, runOptions=RunOptions.ON_CHANGING)
        self.registerFunction(self.updateOnChanged, runOptions=RunOptions.ON_CHANGED)
        self.registerFunction(
            self.updateOnBtnOrChanged,
            runOptions=[
                RunOptions.ON_BUTTON,
                RunOptions.ON_CHANGING,
                RunOptions.ON_CHANGED,
            ],
        )
        self.registerFunction(
            self.updateOnBtnOrShortcut,
            runActionTemplate={"shortcut": "Ctrl+C"},
            runOptions=RunOptions.ON_BUTTON,
        )

    @printResult
    @bind(a=dict(type="slider", limits=[0, 15], step=0.15))
    def updateOnChanging(self, a=5, b=6, c="string"):
        LAST_RESULT.value = f"a={a}, b={b}, c={c}"

    @printResult
    def updateOnBtnOrShortcut(self, boolOp=False):
        LAST_RESULT.value = str(boolOp)

    @printResult
    @bind(f=dict(limits=[0, 17], step=3), g=dict(type="text"))
    def updateOnChanged(self, f=5, g="six\nseven"):
        LAST_RESULT.value = f"f={f}, g={g}"

    @printResult
    @bind(lstOp=dict(type="list", limits=["five", "six", "seven"]))
    def updateOnBtnOrChanged(self, lstOp="five"):
        LAST_RESULT.value = lstOp

    @printResult
    @bind(fileOp=dict(type="file"))
    def updateOnApply(self, fileOp="./"):
        """
        Parameters
        ----------
        fileOp
            File picker can either select existing or non-existing files, folders,
            or multiple files depending on specified configurations.
        """
        LAST_RESULT.value = fileOp


def run():
    app = pg.mkQApp()

    editor = MyClass(directory="./")
    editor.show()
    app.exec_()


if __name__ == "__main__":
    run()



import typing as t

from pyqtgraph.Qt import QtCore, QtGui, QtWidgets


class StringListValidator(QtGui.QValidator):
    def __init__(
        self,
        parent=None,
        strList: t.Sequence[str] = None,
        model: QtCore.QStringListModel = None,
        validateCase=False,
    ) -> None:
        super().__init__(parent)
        self.validateCase = validateCase
        self.strList = strList
        self.model = model

    def validate(self, input_: str, pos: int):
        if self.model:
            self.strList = [
                self.model.index(ii, 0).data() for ii in range(self.model.rowCount())
            ]
        strList = cmpStrList = [input_] if self.strList is None else self.strList
        cmpInput = input_
        if not self.validateCase:
            cmpInput = input_.lower()
            cmpStrList = [s.lower() for s in strList]

        try:
            matchIdx = cmpStrList.index(cmpInput)
            input_ = strList[matchIdx]
            state = self.State.Acceptable
        except ValueError:
            if any(cmpInput in str_ for str_ in cmpStrList):
                state = self.State.Intermediate
            else:
                state = self.State.Invalid
        return state, input_, pos


class PopupLineEditor(QtWidgets.QLineEdit):
    def __init__(
        self,
        parent: QtWidgets.QWidget = None,
        model: QtCore.QAbstractItemModel = None,
        placeholderText="Press Tab or type...",
        clearOnComplete=True,
        forceMatch=True,
        validateCase=False,
    ):
        super().__init__(parent)
        self.setPlaceholderText(placeholderText)
        self.clearOnComplete = clearOnComplete
        self.forceMatch = forceMatch
        self.validateCase = validateCase
        self.model: QtCore.QAbstractListModel = QtCore.QStringListModel()
        if model is None:
            model = self.model

        self.vdator = StringListValidator(parent=self, validateCase=validateCase)
        self.setValidator(self.vdator)

        self.setModel(model)

    def setModel(self, model: QtCore.QAbstractListModel):
        completer = QtWidgets.QCompleter(model, self)
        completer.setCaseSensitivity(QtCore.Qt.CaseSensitivity.CaseInsensitive)
        completer.setCompletionRole(QtCore.Qt.ItemDataRole.DisplayRole)
        completer.setFilterMode(QtCore.Qt.MatchFlag.MatchContains)
        if self.clearOnComplete:
            completer.activated.connect(lambda: QtCore.QTimer.singleShot(0, self.clear))

        self.textChanged.connect(lambda: self.resetCompleterPrefix())

        self.setCompleter(completer)
        self.model = model
        if self.forceMatch:
            self.vdator.model = model

    # TODO: Get working with next prev focusing for smoother logic
    # def focusNextPrevChild(self, nextChild: bool):
    #   if self.forceMatch and self.text() not in self.completer().model().stringList():
    #     dummyFocusEv = QtGui.QFocusEvent(QtCore.QEvent.FocusOut)
    #     self.focusOutEvent(dummyFocusEv)
    #     return False
    #   return super().focusNextPrevChild(nextChild)

    def _chooseNextCompletion(self, incAmt=1):
        completer = self.completer()
        popup = completer.popup()
        if popup.isVisible() and popup.currentIndex().isValid():
            nextIdx = (completer.currentRow() + incAmt) % completer.completionCount()
            completer.setCurrentRow(nextIdx)
        else:
            completer.complete()
        popup.show()
        popup.setCurrentIndex(completer.currentIndex())
        popup.setFocus()

    def event(self, ev: QtCore.QEvent):
        if ev.type() != QtCore.QEvent.Type.KeyPress:
            return super().event(ev)

        ev: QtGui.QKeyEvent
        key = ev.key()
        if key == QtCore.Qt.Key.Key_Tab:
            incAmt = 1
        elif key == QtCore.Qt.Key.Key_Backtab:
            incAmt = -1
        else:
            return super().event(ev)
        self._chooseNextCompletion(incAmt)
        return True

    def focusOutEvent(self, ev: QtGui.QFocusEvent):
        reason = ev.reason()
        if reason in [
            QtCore.Qt.FocusReason.TabFocusReason,
            QtCore.Qt.FocusReason.BacktabFocusReason,
            QtCore.Qt.FocusReason.OtherFocusReason,
        ]:
            # Simulate tabbing through completer options instead of losing focus
            self.setFocus()
            completer = self.completer()
            if completer is None:
                return
            incAmt = 1 if reason == QtCore.Qt.FocusReason.TabFocusReason else -1

            self._chooseNextCompletion(incAmt)
            ev.accept()
            return
        else:
            super().focusOutEvent(ev)

    def clear(self):
        super().clear()

    def resetCompleterPrefix(self):
        if self.text() == "":
            self.completer().setCompletionPrefix("")



from __future__ import annotations

import typing as t

import numpy as np
import pandas as pd
import pyqtgraph as pg
from _warnings import warn
from pyqtgraph import QtCore, QtGui, QtWidgets
from pyqtgraph.graphicsItems.LegendItem import ItemSample
from pyqtgraph.graphicsItems.ScatterPlotItem import drawSymbol
from pyqtgraph.graphicsItems.ViewBox.ViewBoxMenu import ViewBoxMenu

from qtextras import fns
from qtextras._funcparse import bindInteractorOptions as bind
from qtextras.misc import CompositionMixin
from qtextras.params import ParameterEditor, RunOptions
from qtextras.typeoverloads import FilePath
from qtextras.widgets.easywidget import EasyWidget


class ImageViewer(CompositionMixin, pg.PlotWidget):
    sigMouseMoved = QtCore.Signal(object)  # ndarray(int, int) xy pos rel. to image

    def __init__(self, imageSource: np.ndarray = None, **kwargs):
        super().__init__(**kwargs)
        self.pxColorLbl, self.mouseCoordsLbl = None, None
        """
        Set these to QLabels to have up-to-date information about the image coordinates
        under the mouse
        """
        self.toolsEditor = ParameterEditor(name="Region Tools")
        vb = self.getViewBox()
        self.menu: QtWidgets.QMenu = vb.menu
        self.oldVbMenu: ViewBoxMenu = vb.menu
        # Disable default menus
        self.plotItem.ctrlMenu = None
        self.sceneObj.contextMenu = None

        self.setAspectLocked(True)
        vb.invertY()

        # -----
        # IMAGE
        # -----
        self.imageItem = self.exposes(pg.ImageItem())
        self.imageItem.setZValue(-100)
        self.addItem(self.imageItem)
        if imageSource is not None:
            self.setImage(imageSource)

    def mouseMoveEvent(self, ev: QtGui.QMouseEvent):
        super().mouseMoveEvent(ev)
        pos = ev.position() if hasattr(ev, "position") else ev.localPos()
        relpos = self.imageItem.mapFromScene(pos)
        xyCoord = np.array([relpos.x(), relpos.y()], dtype=int)
        if (
            self.imageItem.image is None
            or np.any(xyCoord < 0)
            or np.any(xyCoord > np.array(self.imageItem.image.shape[:2][::-1]) - 1)
        ):
            return
        imgValue = self.imageItem.image[xyCoord[1], xyCoord[0], ...]
        self.updateCursorInfo(xyCoord, imgValue)
        self.sigMouseMoved.emit(xyCoord)

    def updateCursorInfo(self, xyPos: np.ndarray, pxValue: np.ndarray):
        if pxValue is None:
            return
        if self.mouseCoordsLbl is not None:
            self.mouseCoordsLbl.setText(f"Mouse (x,y): {xyPos[0]}, {xyPos[1]}")

        if self.pxColorLbl is None:
            return
        self.pxColorLbl.setText(f"Pixel Color: {pxValue}")
        if self.imageItem.qimage is None:
            return
        imColor = self.imageItem.qimage.pixelColor(*xyPos)
        grayClr = QtGui.qGray(imColor.rgb())
        fontColor = "black" if grayClr > 127 else "white"
        self.pxColorLbl.setStyleSheet(f"background:{imColor.name()}; color:{fontColor}")

    def setImage(
        self, imageSource: t.Union[FilePath, np.ndarray] = None, stripAlpha=False
    ):
        """
        Allows the user to change the main image either from a filepath or array data.
        Files can be any format accepted by QImage -> pg.imageToArray Optionally strips
        the alpha channel from the image, if it exists (i.e. if its shape is MxNx4)
        """
        if isinstance(imageSource, FilePath.__args__):
            # TODO: Handle alpha channel images. For now, discard that data
            qtImg = QtGui.QImage(str(imageSource))
            nchans = qtImg.bitPlaneCount() // 8
            imageSource = pg.imageToArray(qtImg, transpose=False)

            if nchans >= 3:
                chanOrder = [2, 1, 0]
                # Alpha is always added by imageToArray, so chop it off if necessary
                if nchans == 4:
                    chanOrder.append(3)
                imageSource = imageSource[..., chanOrder]
                # imageToArray creates bgr array, turn this into rgb. Also,
                # some algorithms perform better with contiguous arrays, so make sure
                # potential views after this operation remain contiguous Finally,
                # force a copy to avoid crashing after qtImg (which owns the array) is
                # garbage collected
            imageSource = imageSource.copy(order="C")
        if imageSource is None:
            self.imageItem.clear()
        else:
            if stripAlpha and imageSource.ndim > 2:
                imageSource = imageSource[:, :, :3]
            self.imageItem.setImage(imageSource)

    def widgetContainer(self, asMainWin=True, showTools=True, **kwargs):
        """
        Though this is a PlotWidget class, it has a lot of widget children (toolsEditor
        group, buttons) that are not visible when spawning the widget. This is a
        convenience method that creates a new, outer widget from all the graphical
        elements of an EditableImage.

        Parameters
        ----------
        asMainWin
            Whether to return a QMainWindow or QWidget
        showTools
            If `showMainWin` is True, this determines whether to show the tools editor
            with the window
        **kwargs
            Passed to either EasyWidget.buildMainWin or EasyWidget.BuildWidget,
            depending on the value of `asMainWin`
        """
        if asMainWin:
            wid = EasyWidget.buildMainWindow(self._widgetContainerChildren(), **kwargs)
            self.mouseCoordsLbl = QtWidgets.QLabel()
            self.pxColorLbl = QtWidgets.QLabel()
            wid.statusBar().addWidget(self.mouseCoordsLbl)
            wid.statusBar().addWidget(self.pxColorLbl)
            wid.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.toolsEditor)
            if showTools:
                self.toolsEditor.show()
        else:
            wid = EasyWidget.buildWidget(self._widgetContainerChildren(), **kwargs)
        return wid

    def showAndExec(self):
        win = self.widgetContainer(True)
        QtCore.QTimer.singleShot(0, win.showMaximized)
        QtCore.QCoreApplication.instance().exec_()

    def _widgetContainerChildren(self):
        """
        Returns the children that should be added to the container when
        widgetContainer() is called
        """
        return [self]


class CompositorItemSample(ItemSample):
    def paint(self, p: QtGui.QPainter, *args):
        opts = self.item.opts

        if opts.get("antialias"):
            p.setRenderHint(p.Antialiasing)
        symbol = opts.get("symbol", None)
        p.translate(0, 14)
        drawSymbol(
            p, symbol, opts["size"], pg.mkPen(opts["pen"]), pg.mkBrush(opts["brush"])
        )


class CompositorLegend(pg.LegendItem):
    def paint(self, p: QtGui.QPainter, *args):
        br = self.boundingRect()
        p.setPen(self.opts["pen"])
        p.setBrush(self.opts["brush"])
        p.drawRoundedRect(br, 5, 5)


class MaskCompositor(ImageViewer):
    _cachedCmapLimits = None

    def __init__(self, img: np.ndarray = None):
        super().__init__()
        # -----
        # Create properties
        # -----
        self.masksParameter = fns.getParameterChild(
            self.toolsEditor.rootParameter, "Overlays"
        )
        if self._cachedCmapLimits is None:
            # Set at class level to cache for new instances too
            type(self)._cachedCmapLimits = fns.listAllPgColormaps()

        self.legend = CompositorLegend(
            offset=(5, 5), horSpacing=5, brush="ccce", pen="k"
        )
        self.recordDf = pd.DataFrame(
            columns=["name", "item", "opacity", "isMask"]
        ).set_index("name")
        # Initialized in clearOverlays and updateLabelMap. Given here to avoid
        # intellisense warnings
        self.inUseLabelValues = (
            self.scatterSer
        ) = self.backgroundValues = self.labelToNameMapping = None
        self.updateLabelMap()
        self.legendFontArgs = {"size": "11pt", "color": "k", "bold": True}
        self.curZ = 2

        # -----
        # Configure relationships
        # -----
        self.legend.setParentItem(self.plotItem)
        self.viewbox: pg.ViewBox = self.getViewBox()

        self.allVisible = True
        # for ax in 'left', 'bottom', 'right', 'top':
        #   self.mainImage.plotItem.hideAxis(ax)
        # self.mainImage.setContentsMargins(0, 0, 0, 0)

        def newFmt(name):
            if name.startswith("set"):
                name = name.replace("set", "")
            name = fns.pascalCaseToTitle(name)
            return name

        with fns.nameFormatter.set(newFmt):
            self.propertiesProc = self.toolsEditor.registerFunction(
                self.setOverlayProperties,
                runOptions=RunOptions.ON_CHANGED,
                colormap=dict(limits=self._cachedCmapLimits),
                fullLabelMapColors=dict(ignore=True),
            )
            self.toolsEditor.registerFunction(self.save)
            self.toolsEditor.registerFunction(
                lambda: self.imageItem.setVisible(not self.imageItem.isVisible()),
                name="Toggle Image Visible",
            )
            self.toolsEditor.registerFunction(
                lambda: self.legend.setVisible(not self.legend.isVisible()),
                name="Toggle Legend Visible",
            )
            self.toolsEditor.registerFunction(
                self.toggleAllVisible, parent=("Overlays",)
            )
        if img is not None:
            self.setImage(img)
        self.clearOverlays()

    @staticmethod
    def _createLabelToNameMapping(data=None):
        """
        Helper method to ensure proper dtypes when making a legend series. Helps avoid
        pitfalls with empty data, etc.
        """
        if data is None:
            data = []
        ser = pd.Series(data, dtype=str)
        ser.index = ser.index.astype("int64", copy=False)
        return ser

    def _addItemCtrls(self, record: pd.Series):
        item = record["item"]

        def maskOpts(_, visible=True):
            item.setVisible(visible)

        newParam = fns.getParameterChild(
            self.masksParameter,
            childOpts={"name": record.name, "type": "bool", "value": True},
        )
        newParam.sigValueChanged.connect(maskOpts)

    def setBaseImage(self, baseImg: np.ndarray, clearOverlays=True):
        # self.winSz = baseImg.shape[:2][::-1]
        # self.viewbox.setRange(
        #   xRange=[0, imageItem.shape[1]], yRange=[0, imageItem.shape[0]], padding=0
        # )
        self.setImage(baseImg)
        # self.refreshWinContents()
        if clearOverlays:
            self.clearOverlays()

    def setImage(self, *args, **kwargs):
        super().setImage(*args, **kwargs)
        self._updateLegendPos()

    def _addRecord(self, rec: pd.Series, update=True):
        self.recordDf = pd.concat([self.recordDf, rec.to_frame().T], axis=0)
        self.curZ += 1

        self._addItemCtrls(rec)

        if update:
            self._updateGraphics()

    def addImageItem(self, item: pg.ImageItem, **kwargs):
        kwargs.setdefault("opacity", -1)
        name = self._getUniqueName(kwargs.pop("name", None))
        update = kwargs.pop("update", True)
        kwargs.setdefault("isMask", False)
        newRecord = dict(item=item, **kwargs)
        item.setZValue(self.curZ + 1)
        self.viewbox.addItem(item)

        rec = pd.Series(newRecord, name=name)
        self._addRecord(rec, update)

    def _getUniqueName(self, baseName: str = None):
        if baseName is None:
            baseName = "[No Name]"
        ii = 2
        name = baseName
        while name in self.recordDf.index:
            name = f"{baseName} {ii}"
            ii += 1
        return name

    def addLabelMask(
        self,
        labelMask: np.ndarray,
        name=None,
        clearOverlays=False,
        fullLabelMapInLegend=False,
        **overlayKwargs,
    ):
        """
        Splits a label mask into its constituent labels and adds a mask for each unique
        label.

        Parameters
        ----------
        labelMask
            Grayscale label mask of integers
        name
            Name to identify this mask in the property editor
        clearOverlays
            If *True*, `clearOverlays` will be called before adding this item. This
            is useful if only one mask is viewed at a time
        fullLabelMapInLegend
            See ``updateLegendEntries`` -- this parameter is passed during legend
            creation
        **overlayKwargs
            Keyword arguments passed to setOverlayProperties
        """
        if clearOverlays and len(self.recordDf):
            # Reuse item if possible
            item = self._reuseLastOverlay(labelMask, name)
        else:
            item = None
        maskValues = np.unique(labelMask.ravel())
        # Booleans don't mix well with int series indexing/logic
        # Floats are completely wonky
        if np.issubdtype(maskValues.dtype, np.bool_):
            maskValues = maskValues.astype("uint8")
        elif not np.issubdtype(maskValues.dtype, np.integer):
            raise ValueError(
                f"Overlays can only be created using integer dtypes. Encountered"
                f" `{maskValues.dtype}`"
            )

        self.inUseLabelValues = np.union1d(self.inUseLabelValues, maskValues)
        self.updateLegendEntries(useFullLabelMap=fullLabelMapInLegend)
        if item is None:
            self.addImageItem(
                pg.ImageItem(labelMask),
                name=name,
                isMask=True,
                opacity=-1,
                update=False,
            )
        self.propertiesProc(**overlayKwargs)

    def _reuseLastOverlay(self, newImageData=None, newName=None, isMask=True):
        """
        When a set of images must be overlaid separately, it is more efficient to reuse
        an image item compared to constantly adding and removing them from a scene.
        Assumes at least one record is already present in ``self.recordDf``
        """
        self.clearOverlays(self.recordDf.index[:-1])
        # Use two steps to update name so conflicts are avoided
        self.recordDf.index = [None]
        self.recordDf.index = [self._getUniqueName(newName)]
        rec = self.recordDf.iloc[-1]
        rec["isMask"] = isMask
        item = rec["item"]
        item.setImage(newImageData)
        return item

    def setLegendFontStyle(self, startItemIdx=0, **lblTxtArgs):
        for item in self.legend.items[startItemIdx:]:
            for single_item in item:
                if isinstance(single_item, pg.LabelItem):
                    single_item.setText(single_item.text, **lblTxtArgs)

    def updateLegendEntries(self, useFullLabelMap=False, updateOverlays=False):
        """
        Parameters
        ----------
        useFullLabelMap
            If *True*, guarantees that every entry in the label->name mapping will be
            present in the legend.
        updateOverlays
            If *True*, the image overlay properties will be refereshed
        """
        if useFullLabelMap:
            labels = np.union1d(self.inUseLabelValues, self.labelToNameMapping.index)
        else:
            labels = self.inUseLabelValues
        labels = np.setdiff1d(labels, self.backgroundValues)
        names = self.labelToNameMapping.reindex(labels)
        needsEntry = names.isna()
        names[needsEntry] = names.index[needsEntry].map(str)
        self.legend.clear()
        self.scatterSer = pd.Series(
            [None] * len(names), index=names.index, dtype=object
        )
        # Make sure alpha is not carried over
        for value, name in names.items():
            scat = pg.ScatterPlotItem(symbol="s", width=5)
            self.legend.addItem(CompositorItemSample(scat), name=name)
            self.scatterSer.at[value] = scat
        if updateOverlays:
            self.propertiesProc()

    def updateLabelMap(
        self,
        labelMap: dict | pd.Series = None,
        backgroundValues=0,
        updateOverlays=False,
    ):
        """
        Parameters
        ----------
        labelMap
            Series or dict-like with integer key and string value. Gives a legend name
            to each label in the image. Unspecified labels will be named by their
            numeric value, e.g. 34 will be '34'. Note that this will update the global
            legend, meaning if {10 -> 'test'} is a legend entry, it will overwrite the
            old legend value for 10.
        backgroundValues
            Represents label mask values that should be transparent in the overlay
        updateOverlays
            If *True*, overlay graphics will be refreshed to reflect these updates. Keep
            *False* if several updates will happen in sequence.
        """
        if labelMap is None:
            labelMap = {}
        if np.isscalar(backgroundValues):
            backgroundValues = [backgroundValues]
        self.backgroundValues = np.array(backgroundValues)
        self.labelToNameMapping = self._createLabelToNameMapping(labelMap)
        if updateOverlays:
            self._updateGraphics()

    @bind(opacity=dict(limits=[0, 1], step=0.1), colormap=dict(type="popuplineeditor"))
    def setOverlayProperties(
        self, colormap="magma", opacity=0.6, fullLabelMapColors=False
    ):
        """
        Sets overlay properties

        Parameters
        ----------
        colormap
            Colormap to use for the overlay. Should be a string that names a pyqtgraph
            colormap
        opacity
            Opacity of the overlay, from 0 (transparent) to 1 (opaque)
        fullLabelMapColors
            If *True*, enough colors will be generated for all labels in
            ``self.labelToNameMapping``. If *False*, only enough colors will be
            generated for the labels in the added masks.
        """
        maskIdxs = self.recordDf["isMask"].to_numpy(bool)
        cmap = fns.getAnyPgColormap(colormap)
        labelValues = self.scatterSer.index.values
        if fullLabelMapColors:
            labelValues = np.union1d(labelValues, self.labelToNameMapping.index.values)
        if not len(labelValues):
            pass  # return
        if cmap is None:
            raise ValueError(f"Invalid colormap: {colormap}")
        colors = cmap.getLookupTable(nPts=len(labelValues), alpha=True)
        colors[:, -1] = opacity * 255
        # If a labelmap has entries, try to scale the colormap to contain all
        # possibilities
        numEntries = (
            max(
                np.max(labelValues, initial=-1),
                np.max(self.labelToNameMapping.index.values, initial=-1),
            )
            + 1
        )
        lut = np.zeros((numEntries, 4), dtype="uint8")
        lut[labelValues] = colors
        bgValsInLut = self.backgroundValues[self.backgroundValues < len(lut)]
        lut[bgValsInLut] = 0
        for idx, scat in self.scatterSer.items():
            # Make sure alpha is not carried over
            color = lut[idx, :-1]
            brush, pen = pg.mkBrush(color), pg.mkPen(color)
            scat.setData(symbol="s", brush=brush, pen=pen, width=5)
        for _, maskRec in self.recordDf[maskIdxs].iterrows():
            itemOpacity = maskRec["opacity"]
            if itemOpacity < 0:
                itemOpacity = opacity
            maskRec["item"].setOpts(
                lut=lut, opacity=itemOpacity, levels=[0, numEntries - 1]
            )
        # Handle non-mask items who still want controlled opacity
        for _, rec in self.recordDf[
            (~maskIdxs) & self.recordDf["opacity"] < 0
        ].iterrows():
            rec["item"].setOpts(opacity=opacity)
        self.setLegendFontStyle(**self.legendFontArgs)
        self.legend.update()

        # Handle all non-mask items now
        for name, remainingRec in self.recordDf[~maskIdxs].iterrows():
            curOpacity = remainingRec["opacity"]
            if curOpacity < 0:
                curOpacity = opacity
            remainingRec["item"].setOpacity(curOpacity)

    def _updateLegendPos(self):
        imPos = self.imageItem.mapToScene(self.imageItem.pos())
        self.legend.autoAnchor(imPos)

    def _updateGraphics(self):
        self.propertiesProc()

    def clearOverlays(self, clearIdxs=None):
        """
        Clears image item overlays by index. If No index is specified, all overlays are
        removed

        Parameters
        ----------
        clearIdxs
            Indexes of overlays to remove
        """
        if clearIdxs is None:
            clearIdxs = self.recordDf.index
        for imgItem in self.recordDf.loc[clearIdxs, "item"]:
            self.viewbox.removeItem(imgItem)
        keepIdxs = np.setdiff1d(self.recordDf.index, clearIdxs)
        self.recordDf: pd.DataFrame = self.recordDf.loc[keepIdxs].copy()
        self.inUseLabelValues = np.array([], "int64")
        self.updateLegendEntries()
        overlayParam = self.toolsEditor.rootParameter.child("Overlays")

        for idx in clearIdxs:
            overlayParam.child(idx).remove()

    def toggleAllVisible(self):
        for param in self.masksParameter:
            if param.type() == "bool":
                param.setValue(not self.allVisible)
        self.allVisible = not self.allVisible

    @bind(saveFile=dict(type="file", fileMode="AnyFile", acceptMode="AcceptSave"))
    def save(
        self,
        saveFile: FilePath = "",
        cropToViewbox=False,
        toClipboard=False,
        floatLegend=False,
    ):
        """
        Parameters
        ----------
        saveFile
            Save destination. If blank, no file is created.
        toClipboard
            Whether to copy to clipboard
        cropToViewbox
            Whether to only save the visible portion of the image
        floatLegend
            Whether to ancor the legend in the top-left corner (if *False*) or put it
            exactly where it is positioned currently (if *True*). In the latter case, it
            may not appear if out of view and `cropToViewbox` is *True*.
        """
        if self.imageItem.isVisible():
            saveImg = self.imageItem.getPixmap()
        else:
            saveImg = QtGui.QPixmap(*self.imageItem.image.shape[:2][::-1])
        painter = QtGui.QPainter(saveImg)

        visibleMasks = [
            p.name() for p in fns.flattenedParameters(self.masksParameter) if p.value()
        ]
        for name, item in self.recordDf["item"].items():
            if name in visibleMasks:
                painter.setOpacity(item.opacity())
                item.paint(painter)
        painter.setOpacity(1.0)

        if cropToViewbox:
            maxBnd = np.array(self.image.shape[:2][::-1]).reshape(-1, 1) - 1
            vRange = np.array(self.viewbox.viewRange()).astype(int)
            vRange = np.clip(vRange, 0, maxBnd)
            # Convert range to topleft->bottomright
            pts = QtCore.QPoint(*vRange[:, 0]), QtCore.QPoint(*vRange[:, 1])
            # Qt doesn't do so well overwriting the original reference
            origRef = saveImg
            saveImg = origRef.copy(QtCore.QRect(*pts))
            painter.end()
            painter = QtGui.QPainter(saveImg)
        if self.legend.isVisible():
            # Wait to paint legend until here in case cropping is active
            self._paintLegend(painter, floatLegend)

        if toClipboard:
            QtWidgets.QApplication.clipboard().setImage(saveImg.toImage())

        saveFile = str(saveFile)
        if saveFile and not saveImg.save(str(saveFile)):
            warn("Image compositor save failed", UserWarning)
        return saveImg

    def _paintLegend(self, painter: QtGui.QPainter, floatLegend=False):
        oldPos = self.legend.pos()
        try:
            exportScene = pg.GraphicsScene(parent=None)
            oldScale = max(
                np.clip(0.5 / np.array(self.imageItem.pixelSize()), 1, np.inf)
            )
            exportScene.addItem(self.legend)
            self.legend.setScale(oldScale)
            # Legend does not scale itself, handle this
            painter.save()
            if floatLegend:
                viewPos = self.viewbox.viewRect()
                viewX = max(viewPos.x(), 0)
                viewY = max(viewPos.y(), 0)
                imPos = self.legend.mapRectToItem(self.imageItem, self.legend.rect())
                painter.translate(int(imPos.x() - viewX), int(imPos.y() - viewY))
            exportScene.render(painter)
            self.legend.setScale(1 / oldScale)
            painter.restore()
        finally:
            self.legend.setParentItem(self.plotItem)
            self.legend.autoAnchor(oldPos, relative=False)

    def __getstate__(self):
        ret = dict(
            image=self.imageItem.image,
            legendVisible=self.legend.isVisible(),
        )
        if len(self.recordDf):
            ret["recordDf"] = self.recordDf

    def __setstate__(self, state):
        self.__init__(state["image"])
        self.legend.setVisible(state["legendVisible"])
        if "recordDf" in state:
            for _, rec in state["recordDf"].items():
                self._addRecord(rec, False)
            self._updateGraphics()



import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets

import qtextras.widgets.easywidget
from qtextras import ParameterEditor, widgets


def dockgrouping(directory):
    pes = [ParameterEditor(name=str(ii)) for ii in range(3)]
    # grp = ParameterEditorDockGroup(pes, "Test Group")
    window = qtextras.widgets.easywidget.EasyWidget.buildMainWindow(
        [QtWidgets.QTextEdit(), QtWidgets.QPushButton("Test")]
    )
    # window.addDockWidget(QtCore.Qt.DockWidgetArea.RightDockWidgetArea, grp)
    pe = ParameterEditor(name="test", directory=directory)

    def myfunc(a=1):
        print("Test", a)
        return a

    pe.registerFunction(
        myfunc, runOptions="action", runActionTemplate=dict(shortcut="Ctrl+C")
    )
    pes.append(pe)
    for pe in pes:
        pe.createWindowDock(window)

    return window

def test_dockgrouping(tmp_path):
    assert dockgrouping(tmp_path)

if __name__ == "__main__":
    pg.mkQApp()
    win = dockgrouping(".")
    win.raise_()
    win.show()
    pg.exec()



import ast
import configparser
import copy
import inspect
import textwrap
import typing as t
from importlib.util import find_spec
from io import StringIO

from pyqtgraph.parametertree import (
    Interactor as PtreeInteractor,
    RunOptions as PtreeRunOpts,
)
from pyqtgraph.parametertree.interactive import PARAM_UNSET
from pyqtgraph.parametertree.Parameter import PARAM_TYPES

from qtextras import fns

T = t.TypeVar("T")


class FROM_PREV_IO(PARAM_UNSET):
    """
    Helper class to indicate whether a key in this IO is supposed to come from
    a previous process stage. Typical usage:
    ```if not hyperparam: self[k] = self.FROM_PREV_IO```

    Typically, process objects will have two IO dictionaries: One that hold the input
    spec (which makes use of `FROM_PREV_IO`) and one that holds the runtime process
    values. The latter IO will not make use of `FROM_PREV_IO`.
    """


class RunOptions(PtreeRunOpts):
    # Backwards compatibility
    ON_BUTTON = PtreeRunOpts.ON_ACTION


def bindInteractorOptions(overwrite=False, **opts):
    """
    Decorator to bind options to an interactor. This is useful for when the
    call to `interact` is separate from the function definition, but parameter
    options should be located with the function definition. Used as a decorator:
    >>> from qtextras import bindInteractorOptions as bind, QtExtrasInteractor
    >>> @bind(a=dict(step=5))
    ... def myFunc(a=1):
    ...     print(a)
    >>> step = QtExtrasInteractor().interact(myFunc).child("a").opts["step"]
    >>> assert step == 5
    """

    def wrapper(func: T) -> T:
        if hasattr(func, "__interactor_bind_options__") and not overwrite:
            func.__interactor_bind_options__.update(opts)
        else:
            func.__interactor_bind_options__ = opts
        return func

    return wrapper


class QtExtrasInteractor(PtreeInteractor):
    titleFormat = fns.nameFormatter
    displayIgnored = False

    def functionToParameterDict(self, function, **overrides):
        if hasattr(function, "__interactor_bind_options__"):
            newOverrides = function.__interactor_bind_options__
            if overrides:
                newOverrides = copy.deepcopy(newOverrides)
            for kk, vv in overrides.items():
                if not isinstance(vv, dict):
                    overrides[kk] = dict(value=vv)
            overrides = fns.hierarchicalUpdate(newOverrides, overrides)
        ret = super().functionToParameterDict(function, **overrides)
        for ch in ret.get("children", []):
            if ch["value"] is PARAM_UNSET:
                # Avoid causing failures from unspecified parameters, since qtextras
                # allows chaining multiple processes together
                ch["value"] = FROM_PREV_IO
        # By default, don't nest if no children are present
        return ret

    def resolveAndHookupParameterChild(
        self, functionGroup, childOpts, interactiveFunction
    ):
        extra = interactiveFunction.extra
        if childOpts["value"] is FROM_PREV_IO and childOpts["name"] not in extra:
            extra[childOpts["name"]] = FROM_PREV_IO
        if childOpts["name"] in extra or childOpts.get("ignore"):
            return None
        if childOpts.get("type") not in PARAM_TYPES:
            childOpts["type"] = "display"
            childOpts["value"] = extra.get(childOpts["name"], childOpts["value"])
            if not self.displayIgnored:
                # Avoid placing this child in the tree
                functionGroup = None
        return super().resolveAndHookupParameterChild(
            functionGroup, childOpts, interactiveFunction
        )

    def _resolveFunctionGroup(self, functionDict, interactiveFunction):
        # Possible for name to be overridden at the interactive level without
        # affecting the raw function name. In this case, prefer the interactive
        # name. Note in all cases where `interactiveFunc.__name__ == function.__name__`,
        # this is basically an inexpensive no-op.
        functionDict["name"] = interactiveFunction.__name__
        if self.titleFormat is not None:
            functionDict["title"] = self._nameToTitle(
                functionDict["name"], forwardStringTitle=True
            )

        return super()._resolveFunctionGroup(functionDict, interactiveFunction)


class ParameterlessInteractor(QtExtrasInteractor):
    """
    Places every key in "extra" instead of making parameters
    """

    def interact(self, function, *, ignores=None, **kwargs):
        if ignores is None:
            ignores = []
        ignores = set(ignores)
        funcDict = self.functionToParameterDict(function, **kwargs.get("overrides", {}))
        ignores.update(ch["name"] for ch in funcDict["children"])
        return super().interact(function, ignores=ignores, **kwargs)


class _DocstringInteractor(QtExtrasInteractor):
    streamFormat = "ini"
    titleFormat = fns.nameFormatter
    runActionTemplate = dict(defaultName="Run", type="shortcut")

    def functionToParameterDict(self, function, **overrides):
        docDict = docParser(function.__doc__, self.streamFormat)
        for kk in overrides:
            overrideVal = overrides[kk]
            if kk in docDict:
                if isinstance(overrideVal, dict):
                    docDict[kk].update(overrideVal)
                else:
                    docDict[kk]["value"] = overrideVal
            else:
                docDict[kk] = overrideVal
        docDict.pop("tip", None)
        return super().functionToParameterDict(function, **docDict)


def docParser(docstring: str, streamFormat="ini"):
    """
    From a function docstring, extracts relevant information for how to create smarter
      parameter boxes.

    Parameters
    ----------
    docstring
        Function docstring
    streamFormat
        Format of the docstring, either yaml or ini
    """
    if streamFormat == "yaml":
        parser = parseYamlDocstring
    elif streamFormat == "ini":
        parser = parseIniDocstring
    else:
        raise ValueError(f"Unknown docstring stream format: {streamFormat}")
    return parser(docstring)


def parseYamlDocstring(doc: str):
    out = {}
    parsed = dp.parse(doc)
    out["tip"] = "\n".join(
        [
            desc
            for desc in [parsed.short_description, parsed.long_description]
            if desc is not None
        ]
    )
    for param in parsed.params:
        stream = StringIO(param.description)
        try:
            paramDoc = fns.yamlLoad(stream)
        except fns.YAMLError:
            # Some problem parsing, treat as string
            paramDoc = param.description
        if isinstance(paramDoc, str):
            paramDoc = {"tip": paramDoc}
        if paramDoc is None:
            continue
        out[param.arg_name] = dict(name=param.arg_name, **paramDoc)
    return out


def parseIniDocstring(doc: str):
    """
    Parses function documentation for relevant parameter definitions. `doc` must be
    formatted like an .ini file, where each option's parameters are preceded by a
    ``[<arg>.options]`` header. See the examples in tests/parametertree/test_docparser
    for valid configurations. Note that if the `docstring_parser` module is available
    in the python environment, section headers as described above are *not required*
    since they can be inferred from the properly formatted docstring. The return value
    is a dict where each entry contains the found specifications of a given argument,
    i.e {"param1": {"limits": [0, 10], "title": "hi"}, "param2": {...}}, etc. Note that
    currently only literal evaluation with builtin objects is supported, i.e. the
    return result of ast.literal_eval on the value string of each option.

    Parameters
    ----------
    doc
      Documentation to parse
    """
    # Use docstring parser if it's available, otherwise basic parsing
    if find_spec("docstring_parser") is not None:
        return _parseIniDocstring_docstringParser(doc)
    else:
        return _parseIniDocstring_basic(doc)


def _parseIniDocstring_docstringParser(doc):
    """
    Use docstring_parser for a smarter version of the ini parser. Doesn't require
    ``[<arg>.options]`` headers and can handle more dynamic parsing cases
    """
    # Revert to basic method if ini headers are already present
    if not doc or ".options]\n" in doc:
        return _parseIniDocstring_basic(doc)
    import docstring_parser

    out = {}
    parsed = docstring_parser.parse(doc)
    out["tip"] = "\n".join(
        [
            desc
            for desc in [parsed.short_description, parsed.long_description]
            if desc is not None
        ]
    )
    for param in parsed.params:
        # Construct mini ini file around each parameter
        header = f"[{param.arg_name}.options]"
        miniDoc = param.description
        if header not in miniDoc:
            miniDoc = f"{header}\n{miniDoc}"
        # top-level parameter no longer represents whole function
        update = _parseIniDocstring_basic(miniDoc)
        update.pop("tip", None)
        out.update(update)
    return out


def _parseIniDocstring_basic(doc):
    """
    Implements quite primitive parsing of parameter signatures. Looks for
    ``[<param-name>.options]`` blocks in the config string and parses them with a
    ConfigParser. Note that all ``*.options`` blocks are expected to start at the
    same indentation level. Options shared across all parameters can be specified under
    a ``[DEFAULT]`` block.
    """
    out = {}
    # Adding "[DEFAULT] to the beginning of the doc will consume non-parameter
    # descriptions Account for things in commonly supported docstring formats:
    fmtDoc = _formatOptionsFromDoc(doc or "")

    parser = configparser.ConfigParser(allow_no_value=True)
    # Save case sensitivity
    parser.optionxform = str
    try:
        parser.read_string(fmtDoc)
    except configparser.Error:
        # Many things can go wrong reading a badly-formatted docstring, so failsafe by
        # returning early
        return out
    defaultKeys = set(parser.defaults())
    for kk, vv in parser.items():
        if not kk.endswith(".options"):
            continue
        paramName = kk.split(".")[0]
        # vv is a section with options for the parameter, but each option must be
        # literal eval'd for non-string values
        paramValues = dict(vv)
        # Consolidate all non-valued key strings into a single tip since they likely
        # belong to the argument documentation. Since dict preserves order, it should
        # come back out in the right order.
        backupTip = ""
        for paramK, paramV in list(paramValues.items()):
            if paramV is None:
                # Considered a tip of the current option if no corresponding value
                # and not a default option
                if paramK not in defaultKeys:
                    backupTip = f"{backupTip} {paramK}"
                # Remove this from the return value since it isn't really meaningful
                del paramValues[paramK]
                continue
            # noinspection PyBroadException
            try:
                paramValues[paramK] = ast.literal_eval(paramV)
            except Exception:
                # There are many reasons this can fail, a safe fallback is the original
                # string value
                pass
        if backupTip:
            paramValues.setdefault("tip", backupTip.strip())
        out[paramName] = paramValues
    # Since function documentation can be used as a description for whatever group
    # parameter hosts these parameters, store it in a name guaranteed not to collide
    # with parameter names since its invalid variable syntax (contains '-')
    out["tip"] = "\n".join(parser.defaults())
    return out


def _stripNSpaces(string, n):
    for ii in range(min(n, len(string))):
        if not string[ii].isspace():
            break
    else:
        ii = 0
    return string[ii:]


def _countStartingSpaces(line):
    ii = 0
    for ii in range(len(line)):
        if line[ii].isspace():
            ii += 1
    return ii


def _formatOptionsFromDoc(doc):
    """
    Dedents doc to [*.options] level for ConfigParser parsing and removes leading ":"
    from each line if needed. These are both factors that complicate config parsing
    later on
    """
    dedent = None
    docLines = doc.splitlines()
    defaultLine = -1
    for ii, line in enumerate(docLines):
        if dedent is None and ".options]" in line:
            dedent = _countStartingSpaces(line)
        stripped = line.strip()
        if stripped.startswith(":"):
            # :param: style documentation violates ini standards
            docLines[ii] = line.replace(":", "", 1)
        elif stripped == "[DEFAULT]":
            defaultLine = ii
    if defaultLine > -1:
        # Remove all lines before default if it's present, since they can't be parsed
        docLines = docLines[defaultLine:]
    if dedent is not None:
        doc = "\n".join(_stripNSpaces(line, dedent) for line in docLines)
    else:
        doc = textwrap.dedent(doc)

    # Now that wrapping ensures lines start without whitespace relative to options,
    # default block can be added if needed
    if defaultLine < 0:
        doc = f"[DEFAULT]\n{doc}"

    return doc



from __future__ import annotations

import html
import logging
import sys
import typing as t
import weakref
from textwrap import wrap
from traceback import format_exception

import numpy as np
from pyqtgraph import QtCore, QtWidgets

from qtextras import fns


class ScrollableMessageDialog(QtWidgets.QDialog):
    def __init__(
        self,
        parent: QtWidgets.QWidget = None,
        messageType="Information",
        msg="",
        detailedMsg="",
    ):
        super().__init__(parent)
        style = self.style()
        self.setModal(True)

        styleIcon = getattr(
            QtWidgets.QStyle.StandardPixmap, f"SP_MessageBox{messageType}"
        )
        self.setWindowTitle(messageType)
        self.setWindowIcon(style.standardIcon(styleIcon))

        verticalLayout = QtWidgets.QVBoxLayout(self)

        scrollArea = QtWidgets.QScrollArea(self)
        scrollArea.setWidgetResizable(True)
        scrollAreaWidgetContents = QtWidgets.QWidget()

        scrollLayout = QtWidgets.QVBoxLayout(scrollAreaWidgetContents)

        # Set to message with trace first so sizing is correct
        msgLbl = QtWidgets.QLabel(detailedMsg, scrollAreaWidgetContents)
        msgLbl.setTextInteractionFlags(
            QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
            | QtCore.Qt.TextInteractionFlag.TextSelectableByKeyboard
        )
        msgLbl.setTextFormat(QtCore.Qt.TextFormat.PlainText)
        scrollLayout.addWidget(
            msgLbl,
            0,
            QtCore.Qt.AlignmentFlag.AlignLeft | QtCore.Qt.AlignmentFlag.AlignTop,
        )
        scrollArea.setWidget(scrollAreaWidgetContents)
        verticalLayout.addWidget(scrollArea)

        btnLayout = QtWidgets.QHBoxLayout()
        ok = QtWidgets.QPushButton("Ok", self)
        toggleTrace = QtWidgets.QPushButton("Toggle Details", self)
        btnLayout.addWidget(ok)
        btnLayout.addWidget(toggleTrace)
        spacerItem = QtWidgets.QSpacerItem(
            ok.width(),
            ok.height(),
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum,
        )
        ok.clicked.connect(self.close)

        sh = self.sizeHint()
        newWidth = max(sh.width(), self.width())
        newHeight = max(sh.height(), self.height())
        self.resize(newWidth, newHeight)

        showDetailedMsg = False

        def updateTxt():
            nonlocal showDetailedMsg
            if showDetailedMsg:
                newText = detailedMsg.replace("\n", "<br>")
                msgLbl.setTextFormat(QtCore.Qt.TextFormat.RichText)
            else:
                newLines = msg.splitlines()
                allLines = []
                for line in newLines:
                    if line == "":
                        line = [line]
                    else:
                        line = wrap(line)
                    allLines.extend(line)
                newText = "<br>".join(allLines)
                msgLbl.setTextFormat(QtCore.Qt.TextFormat.RichText)
            showDetailedMsg = not showDetailedMsg
            msgLbl.setText(newText)

        self.msgLbl = msgLbl
        toggleTrace.clicked.connect(lambda: updateTxt())

        btnLayout.addItem(spacerItem)
        verticalLayout.addLayout(btnLayout)
        self.toggleTrace = toggleTrace
        ok.setFocus()
        updateTxt()


class QtAppHandler(logging.Handler):
    _weakWindow: weakref.ReferenceType = None

    def __init__(
        self, level, win: QtWidgets.QMainWindow = None, exceptionsOnly=False
    ) -> None:
        super().__init__(level)
        self.attachToWindow(win)
        self.exceptionsOnly = exceptionsOnly

    def attachToWindow(self, win=None):
        if win is not None:
            win = weakref.ref(win)
        self._weakWindow = win

    def getWindow(self) -> t.Optional[QtWidgets.QMainWindow]:
        if self._weakWindow is None:
            return None
        return self._weakWindow()


class DialogHandler(QtAppHandler):
    def emit(self, record: logging.LogRecord):
        if record.exc_info:
            msg = html.escape(str(record.exc_info[1]))
            excStr = "".join(format_exception(*record.exc_info))
            detailed = html.escape(excStr)
        else:
            msg = super().format(record)
            detailed = getattr(record, "detailed", msg)
        dlgType = "Critical" if record.levelno == logging.CRITICAL else "Information"

        dlg = ScrollableMessageDialog(self.getWindow(), dlgType, msg, detailed)

        def doExec():
            dlg.exec()

        # Using function instead of directly connecting to dlg.exec keeps garbage from
        # being collected just long enough
        QtCore.QTimer.singleShot(0, doExec)

    def filter(self, record: logging.LogRecord):
        """
        Optionally only activates this handler if the message came from an exception
        """
        if not self.exceptionsOnly:
            return super().filter(record)
        try:
            return len(record.exc_info) > 1
        except Exception:
            return False


class TimedMessageHandler(QtAppHandler):
    def __init__(
        self, *args, defaultMsgtimeout_ms=3000, maxLevel=None, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.defaultMsgTimeout = defaultMsgtimeout_ms
        if maxLevel is None:
            maxLevel = sys.maxsize
        self.maxLevel = maxLevel

    def emit(self, record: logging.LogRecord):
        win = self.getWindow()
        if win is None:
            return
        if hasattr(record, "msgTimeout"):
            msgTimeout = record.msgTimeout
        else:
            msgTimeout = self.defaultMsgTimeout
        self.makeNotification(self.format(record), msgTimeout)

    def filter(self, record: logging.LogRecord):
        return record.levelno <= self.maxLevel and logging.Handler.filter(self, record)

    def makeNotification(self, msg: str, timeout_ms: int = None):
        raise NotImplementedError


class StatusBarHandler(TimedMessageHandler):
    def makeNotification(self, msg: str, timeout_ms: int = None):
        self.getWindow().statusBar().showMessage(msg, timeout_ms)


class FadeNotifyHandler(TimedMessageHandler):
    updateFps = 10

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # If references aren't kept to labels, they will caused wrapped c++ deletion
        # errors List instead of single ref since multiple messages can exist at once
        # Dict to also associate animations with their window
        self.msgRefs = {}

    def makeNotification(self, msg: str, timeout_ms: int = None):
        """
        Creates a fading green dialog that shows the message and disappears after
        timeout
        """
        style = """\
    QLabel {
      background-color: #0f0;
      border: 1px solid #000;
      color: #000;
    };
    """
        fadeMsg = QtWidgets.QLabel(msg, self.getWindow())
        fadeMsg.setStyleSheet(style)
        # For some reason, width doesn't set properly here
        fadeMsg.setFixedWidth(fadeMsg.sizeHint().width())
        fadeMsg.setAlignment(QtCore.Qt.AlignCenter)
        fadeMsg.setAttribute(QtCore.Qt.WidgetAttribute.WA_TransparentForMouseEvents)
        parent: QtWidgets.QMainWindow = self.getWindow()
        if parent is not None:
            metrics = fadeMsg.fontMetrics()
            padding = (parent.width() - metrics.width(msg)) // 2
            msgX = padding
            # Make sure multiple messages don't overlap
            msgY = int(0.05 * parent.height())
            msgY += self.getUniqueLblOffset(msgY)
            fadeMsg.move(msgX, msgY)
        anim = self.makeAnimation(fadeMsg, timeout_ms)
        # 10 fps
        anim.start(1000 // self.updateFps)
        fadeMsg.show()

    def getUniqueLblOffset(self, constHeight):
        allYPos = np.array([label.pos().y() - constHeight for label in self.msgRefs])
        if not len(allYPos):
            return 0
        height = next(iter(self.msgRefs)).height()
        offset = 0
        recalc = True
        while recalc:
            offset += 0.8 * height
            recalc = np.min(np.abs(allYPos - offset)) < 0.01
        return offset

    def closeMsgWidget(self, msgWidget):
        """
        Closes the specified messages and deletes corresponding references
        """
        msgWidget.hide()
        msgWidget.deleteLater()
        self.msgRefs[msgWidget].deleteLater()
        del self.msgRefs[msgWidget]

    def makeAnimation(self, widget, timeout_ms):
        """
        The default method of using a PropertyAnimation doesn't repaint the widget,
        so do it a quick and dirty way using timers manually
        """

        def changeOpacity():
            for opacity in np.linspace(1, 0, int(timeout_ms / 1000 * self.updateFps)):
                effect.setOpacity(opacity)
                # widget.update()
                yield
            # Animation done, wait one tick to prevent overlaps
            yield
            self.closeMsgWidget(widget)

        # Don't start right away
        effect = QtWidgets.QGraphicsOpacityEffect(widget)
        widget.setGraphicsEffect(effect)
        anim = fns.timedExec(changeOpacity, interval_ms=0)
        self.msgRefs[widget] = anim
        return anim



from qtextras.widgets.buttoncollection import ButtonCollection
from qtextras.widgets.console import ConsoleWidget, safeSpawnDevConsole
from qtextras.widgets.easywidget import EasyWidget
from qtextras.widgets.imageviewer import ImageViewer, MaskCompositor
from qtextras.widgets.lineeditor import PopupLineEditor, StringListValidator
from qtextras.widgets.loghandlers import (
    DialogHandler,
    FadeNotifyHandler,
    QtAppHandler,
    ScrollableMessageDialog,
    StatusBarHandler,
    TimedMessageHandler,
)



import numpy as np
import pytest

from qtextras.misc import ActionStack, DeferredActionStackMixin as DASM

COUNT = 10


class StackForTesting(ActionStack):
    def __init__(self):
        super().__init__()

        @self.undoable()
        def op(num=None):
            if num is None:
                num = len(self.lst)
            self.lst.append(num)
            yield
            self.lst.pop()

        @self.undoable("recurse")
        def recursiveOp_swapEveryTwo(lst=None):
            if lst is None:
                lst = self.lst
            sz = len(lst)
            if sz < 2:
                pass
            elif sz < 4:
                tmp = lst[0]
                lst[0] = lst[1]
                lst[1] = tmp
            else:
                for rng in slice(sz // 2), slice(sz // 2, sz):
                    lst[rng] = recursiveOp_swapEveryTwo(lst[rng])
            yield lst
            recursiveOp_swapEveryTwo(lst)

        def grpOp():
            with self.group("grouping"):
                for _ in range(COUNT):
                    self.op()

        self.recursive = recursiveOp_swapEveryTwo
        self.grpOp = grpOp
        self.op = op

        self.lst = []

    def clear(self):
        super().clear()
        self.lst.clear()


def test_group():
    stack = StackForTesting()
    stack.grpOp()
    stack.op()
    assert stack.lst == list(range(COUNT + 1))
    stack.undo()
    assert stack.lst == list(range(COUNT))
    stack.undo()
    assert stack.lst == []


def test_nested_doable():
    stack = StackForTesting()

    @stack.undoable("outer op")
    def outer():
        stack.lst.append(str(len(stack.lst)))
        stack.op()
        yield
        stack.lst.pop()
        stack.lst.pop()

    outer()
    assert stack.lst == ["0", 1]
    stack.undo()
    assert stack.lst == []


def test_recursive():
    stack = StackForTesting()

    nextPow2 = int(np.power(2, np.ceil(np.log2(COUNT))))
    stack.lst = list(range(nextPow2))
    origLst = stack.lst.copy()

    stack.recursive()
    swapped = stack.lst.copy()

    stack.undo()
    assert stack.lst == origLst

    stack.redo()
    assert stack.lst == swapped
    assert len(stack.actions) == 1


def test_bad_undo():
    stack = StackForTesting()

    for _ii in range(4):
        stack.op()
    for _ii in range(4):
        stack.undo()
    with pytest.warns(UserWarning):
        stack.undo()


def test_bad_redo():
    stack = StackForTesting()
    with pytest.warns(UserWarning):
        stack.redo()
    stack.op()
    with pytest.warns(UserWarning):
        stack.undo()
        stack.redo()
        stack.redo()


def test_invalidate_redos():
    stack = StackForTesting()

    for ii in range(COUNT):
        stack.op()

    assert stack.lst == list(range(COUNT))
    numEntriesToRemomve = COUNT // 3
    for ii in range(numEntriesToRemomve):
        stack.undo()

    numRemainingEntries = COUNT - numEntriesToRemomve
    assert np.sum([a.treatAsUndo for a in stack.actions]) == numRemainingEntries

    stack.op(1)
    # New action should flush old ones
    with pytest.warns(UserWarning):
        stack.redo()
    stack.undo()
    assert len(stack.actions) == numRemainingEntries + 1
    cmplst = list(range(numRemainingEntries))
    assert stack.lst == cmplst
    stack.redo()
    assert stack.lst == cmplst + [1]


def test_ignore_acts():
    stack = StackForTesting()
    with stack.ignoreActions():
        for _ in range(COUNT):
            stack.op()
    assert len(stack.actions) == 0
    with pytest.warns(UserWarning):
        stack.undo()


def test_max_len():
    stack = StackForTesting()
    cnt = stack.actions.maxlen + 20
    for ii in range(cnt):
        stack.op()

    with pytest.warns(UserWarning):
        for ii in range(cnt):
            stack.undo()


def test_grp_composite():
    stack = StackForTesting()
    with stack.group("outer grp", flushUnusedRedos=True):
        with stack.group("inner grp"):
            stack.grpOp()
            stack.recursive()
        stack.op()
        stack.grpOp()
        stack.op()
    assert len(stack.actions) == 1

    postOpLst = stack.lst.copy()
    stack.undo()
    assert len(stack.lst) == 0
    stack.redo()
    assert postOpLst == stack.lst


def test_savepoint():
    stack = StackForTesting()

    for _ in range(COUNT):
        stack.op()
    stack.setSavepoint()

    stack.op()
    assert stack.changedSinceLastSave

    stack.revertToSavepoint()
    assert not stack.changedSinceLastSave

    stack.op()
    stack.op()

    stack.revertToSavepoint()
    assert stack.lst == list(range(COUNT))


def test_reassign_backward():
    stack = StackForTesting()
    newlst = []

    def newAppend():
        newlst.append(5)

    # Test in forward state
    for _ in range(COUNT):
        stack.op()
        stack.actions[-1].reassignBackward(newAppend)
    for _ in range(COUNT):
        stack.undo()
    assert len(stack.actions) == COUNT
    assert newlst == [5] * COUNT

    # Make sure it works going backward
    for _ in range(COUNT):
        stack.redo()
    assert len(stack.lst) == 2 * COUNT
    for _ in range(COUNT):
        stack.undo()
    assert newlst == [5] * COUNT * 2

    # Make sure doing reassign after undo works too
    stack.clear()
    newlst.clear()
    for _ in range(COUNT):
        stack.op()
    for _ in range(COUNT):
        stack.undo()
    for ii in range(COUNT):
        stack.actions[ii].reassignBackward(newAppend)
    for _ in range(COUNT):
        stack.redo()
    for _ in range(COUNT):
        stack.undo()
    assert newlst == [5] * COUNT


def test_deferred_stack():
    class A(DASM):
        @DASM.undoable("test")
        def test(self, num: int):
            print(num)
            yield
            print(num + 1)

        @DASM.undoable("test 2")
        def moretest(self, x=1, y=3):
            print(dict(x=x, y=y))
            yield
            print(dict(x=x + 1, y=y + 1))

    class B(A):
        pass

    class C(A):
        @B.undoable("this one")
        def test2(self):
            yield "test 2 first"
            return "test 2 second"

    class D(C):
        pass

    a = A()
    with A.setStack(ActionStack()):
        b = B()
        c = C()
    assert a.actionStack is not b.actionStack
    assert b.actionStack is c.actionStack

    with B.setStack(ActionStack()):
        a = A()
        b = B()
        c = C()
        d = D()
    # B.setstack means a (as an earlier class in the hierarchy) should not be set
    # Also, C (and D) didn't come from B, so those shouldn't be overridden either
    assert a.actionStack is not b.actionStack is not c.actionStack is not d.actionStack

    with C.setStack(ActionStack()):
        a = A()
        b = B()
        c = C()
        d = D()
    assert a.actionStack is not b.actionStack is not c.actionStack is d.actionStack

    C.actionStack = mystack = ActionStack()
    d = D()
    assert d.actionStack is mystack
    with C.setStack(ActionStack()):
        c = C()
    # Make sure reset happens correctly
    assert C.actionStack is mystack
    assert c.actionStack is not mystack



from __future__ import annotations

import typing as t

import numpy as np
import pandas as pd
from pyqtgraph import QtCore

from qtextras.constants import PrjEnums


class PandasTableModel(QtCore.QAbstractTableModel):
    """
    Class to populate a table view with a pandas dataframe
    """

    sigDataChanged = QtCore.Signal(object)
    defaultEmitDict = {
        "deleted": np.array([]),
        "changed": np.array([]),
        "added": np.array([]),
    }

    # Will be set in 'changeDefaultRows'
    dataframe: pd.DataFrame
    _defaultSeries: pd.Series

    def __init__(self, defaultSer: pd.Series, parent=None):
        super().__init__(parent)
        self.dataframe = pd.DataFrame()
        self.changeDefaultRows(defaultSer)
        self._nextRowId = 0

    def rowCount(self, parent=None):
        return self.dataframe.shape[0]

    def columnCount(self, parent=None):
        return self.dataframe.shape[1]

    def data(self, index, role=QtCore.Qt.ItemDataRole.DisplayRole):
        if index.isValid():
            value = self.dataframe.iloc[index.row(), index.column()]
            if role == QtCore.Qt.ItemDataRole.DisplayRole:
                return str(value)
            elif role == QtCore.Qt.ItemDataRole.EditRole:
                return value
        return None

    def setData(
        self,
        index: QtCore.QModelIndex,
        value: t.Any,
        role: int = QtCore.Qt.ItemDataRole,
    ) -> bool:
        super().setData(index, role)
        row = index.row()
        col = index.column()
        oldVal = self.dataframe.iat[row, col]
        # Try-catch for case of numpy arrays
        noChange = oldVal == value
        try:
            if noChange:
                return True
        except ValueError:
            # Happens with array comparison
            pass
        self.dataframe.iat[row, col] = value
        self.sigDataChanged.emit()
        return True

    def headerData(self, section, orientation, role=QtCore.Qt.ItemDataRole.DisplayRole):
        if (
            orientation == QtCore.Qt.Orientation.Horizontal
            and role == QtCore.Qt.ItemDataRole.DisplayRole
        ):
            return self.dataframe.columns[section]

    def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlags:
        return QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable

    def addDfRows(
        self, rowData: pd.DataFrame, addType=PrjEnums.ADD_TYPE_NEW, emitChange=True
    ):
        toEmit = self.defaultEmitDict.copy()
        if addType == PrjEnums.ADD_TYPE_NEW:
            # Treat all components as new -> set their IDs to guaranteed new values
            newIds = np.arange(
                self._nextRowId, self._nextRowId + len(rowData), dtype=int
            )
            rowData.set_index(newIds, inplace=True, verify_integrity=False)
            # For new data without all columns, add missing values to ensure they're
            # correctly filled
            if np.setdiff1d(rowData.columns, self.dataframe.columns).size > 0:
                rowData = self.makeDefaultDfRows(len(rowData), rowData)
        else:
            # Merge may have been performed with new components (id -1) mixed in
            needsUpdatedId = rowData.index == -1
            newIds = np.arange(
                self._nextRowId, self._nextRowId + np.sum(needsUpdatedId), dtype=int
            )
            rowData.index[needsUpdatedId] = newIds

        # Merge existing IDs and add new ones
        changedIdxs = np.isin(rowData.index, self.dataframe.index, assume_unique=True)
        changedIds = rowData.index[changedIdxs]
        addedIds = rowData.index[~changedIdxs]

        # Signal to table that rows should change
        self.layoutAboutToBeChanged.emit()
        # Ensure indices overlap with the components these are replacing
        self.dataframe.update(rowData)
        toEmit["changed"] = changedIds

        # Finally, add new components
        compsToAdd = rowData.loc[addedIds]
        self.dataframe = pd.concat((self.dataframe, compsToAdd), sort=False)
        toEmit["added"] = addedIds

        # Retain type information
        self._coerceDfTypes()

        self.layoutChanged.emit()

        self._nextRowId = np.max(self.dataframe.index.to_numpy(), initial=-1) + 1

        if emitChange:
            self.sigDataChanged.emit(toEmit)
        return toEmit

    def removeDfRows(self, idsToRemove: t.Sequence[int] = None, emitChange=True):
        if idsToRemove is None:
            idsToRemove = self.dataframe.index
        toEmit = self.defaultEmitDict.copy()
        # Generate ID list
        existingCompIds = self.dataframe.index
        idsToRemove = np.asarray(idsToRemove)

        # Do nothing for IDs not actually in the existing list
        idsActuallyRemoved = np.isin(idsToRemove, existingCompIds, assume_unique=True)
        if len(idsActuallyRemoved) == 0:
            return toEmit
        idsToRemove = idsToRemove[idsActuallyRemoved]

        tfKeepIdx = np.isin(
            existingCompIds, idsToRemove, assume_unique=True, invert=True
        )

        # Reset manager's component list
        self.layoutAboutToBeChanged.emit()
        self.dataframe = self.dataframe.iloc[tfKeepIdx, :]
        self.layoutChanged.emit()

        # Preserve type information after change
        self._coerceDfTypes()

        # Determine next ID for new components
        self._nextRowId = 0
        if np.any(tfKeepIdx):
            self._nextRowId = np.max(existingCompIds[tfKeepIdx].to_numpy()) + 1

        # Reflect these changes to the component list
        toEmit["deleted"] = idsToRemove
        if emitChange:
            self.sigDataChanged.emit(toEmit)

    def makeDefaultDfRows(self, numRows=1, initData: pd.DataFrame = None):
        """
        Create a dummy table populated with default values from the class default
        pd.Series. If `initData` is provided, it must have rows entries and
        correspond to columns from the default series. these columns will be overridden
        by the init data.
        """
        if numRows == 0:
            return pd.DataFrame(columns=self._defaultSeries.index)
        outDf = pd.DataFrame([self._defaultSeries] * numRows)
        if initData is not None:
            outDf.update(initData.set_index(outDf.index))
        return outDf

    def changeDefaultRows(self, defaultSer: pd.Series):
        self.beginResetModel()
        self._defaultSeries = defaultSer
        self.removeDfRows(self.dataframe.index)
        self.dataframe = self.makeDefaultDfRows(0)
        self.endResetModel()

    def _coerceDfTypes(self):
        """
        Pandas currently has a bug where datatypes are not preserved after update
        operations. Current workaround is to coerce all types to their original values
        after each operation
        """
        for ii, col in enumerate(self.dataframe.columns):
            idealType = type(self._defaultSeries[col])
            if not np.issubdtype(self.dataframe.dtypes[ii], idealType):
                try:
                    self.dataframe[col] = self.dataframe[col].astype(idealType)
                except (TypeError, ValueError):
                    continue



import logging


class PrjEnums:
    ADD_TYPE_MERGE = "merge"
    ADD_TYPE_NEW = "new"

    REG_CREATE_NEW = "new"

    # Logging level for 'attention': Like logging.INFO, but requests more of the
    # user's attention
    LOG_LVL_ATTN = logging.INFO + 1



from __future__ import annotations

from pyqtgraph.parametertree import Parameter, parameterTypes as ptypes
from pyqtgraph.parametertree.Parameter import PARAM_TYPES
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets

from qtextras import fns
from qtextras.widgets.lineeditor import PopupLineEditor

__all__ = [
    "KeySequenceParameterItem",
    "ParameterDialog",
    "PgParameterDelegate",
    "PgPopupDelegate",
    "PopupLineEditorParameterItem",
    "PopupLineEditorParameter",
]


class MonkeyPatchedTextParameterItem(ptypes.TextParameterItem):
    def makeWidget(self):
        textBox: QtWidgets.QTextEdit = super().makeWidget()
        textBox.setTabChangesFocus(True)
        return textBox


# Monkey patch pyqtgraph text box to allow tab changing focus
ptypes.TextParameter.itemClass = MonkeyPatchedTextParameterItem


class ParameterDialog(QtWidgets.QDialog):
    def __init__(self, param: Parameter, parent=None):
        super().__init__(parent)
        self.setModal(True)
        self.param = param
        self.tree = fns.flexibleParameterTree(param)
        self.saveChanges = False

        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.tree)

        okBtn = QtWidgets.QPushButton("Ok")
        cancelBtn = QtWidgets.QPushButton("Cancel")
        btnLay = QtWidgets.QHBoxLayout()
        btnLay.addWidget(okBtn)
        btnLay.addWidget(cancelBtn)
        layout.addLayout(btnLay)

        def okClicked():
            self.saveChanges = True
            self.accept()

        def cancelClicked():
            self.saveChanges = False
            self.reject()

        okBtn.clicked.connect(okClicked)
        cancelBtn.clicked.connect(cancelClicked)


class PgPopupDelegate(QtWidgets.QStyledItemDelegate):
    """
    For pyqtgraph-registered parameter types that don't define an itemClass with
    `makeWidget`, this popup delegate can be used instead which creates a popout
    parameter tree.
    """

    def __init__(self, paramDict: dict, parent=None):
        super().__init__(parent)
        paramDict.setdefault("name", paramDict["type"])
        self.param = Parameter.create(**paramDict)

    def createEditor(self, parent, option, index: QtCore.QModelIndex):
        self.param.setValue(index.data(QtCore.Qt.ItemDataRole.EditRole))
        editor = ParameterDialog(self.param)
        editor.show()
        editor.resize(editor.width() + 50, editor.height() + 30)

        return editor

    def setModelData(
        self,
        editor: QtWidgets.QWidget,
        model: QtCore.QAbstractTableModel,
        index: QtCore.QModelIndex,
    ):
        if editor.saveChanges:
            model.setData(index, editor.param.value())

    def setEditorData(self, editor: QtWidgets.QWidget, index):
        value = index.data(QtCore.Qt.ItemDataRole.EditRole)
        self.param.setValue(value)

    def updateEditorGeometry(
        self,
        editor: QtWidgets.QWidget,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ):
        return


class PgParameterDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parameterDict: dict, parent=None):
        super().__init__(parent)
        errMsg = (
            f"`{self.__class__}` can only create parameter editors from registered"
            f" pyqtgraph widgets whose items subclass `WidgetParameterItem` and are in"
            f" pyqtgraph's `PARAM_TYPES`.\n"
            f"These requirements are not met for type `{parameterDict['type']}`"
        )

        if parameterDict["type"] not in PARAM_TYPES:
            raise TypeError(errMsg)
        parameterDict.update(name="dummy")
        self.param = param = Parameter.create(**parameterDict)
        item = param.makeTreeItem(0)
        if isinstance(item, ptypes.WidgetParameterItem):
            self.item = item
        else:
            raise TypeError(errMsg)

    def createEditor(self, parent, option, index: QtCore.QModelIndex):
        # TODO: Deal with params that go out of scope before yielding a value
        editor = self.item.makeWidget()
        editor.setParent(parent)
        editor.setMaximumSize(option.rect.width(), option.rect.height())
        return editor

    def setModelData(
        self,
        editor: QtWidgets.QWidget,
        model: QtCore.QAbstractTableModel,
        index: QtCore.QModelIndex,
    ):
        model.setData(index, editor.value())

    def setEditorData(self, editor: QtWidgets.QWidget, index):
        value = index.data(QtCore.Qt.ItemDataRole.EditRole)
        editor.setValue(value)

    def updateEditorGeometry(
        self,
        editor: QtWidgets.QWidget,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ):
        editor.setGeometry(option.rect)


class KeySequenceParameterItem(ptypes.WidgetParameterItem):
    """
    Class for creating custom shortcuts. Must be made here since pyqtgraph doesn't
    provide an implementation.
    """

    def makeWidget(self):
        item = QtWidgets.QKeySequenceEdit()

        item.sigChanged = item.editingFinished
        item.value = lambda: item.keySequence().toString()

        def setter(val: QtGui.QKeySequence):
            if val is None or len(val) == 0:
                item.clear()
            else:
                item.setKeySequence(val)

        item.setValue = setter
        self.param.seqEdit = item

        return item

    def updateDisplayLabel(self, value=None):
        # Make sure the key sequence is human-readable
        self.displayLabel.setText(self.widget.keySequence().toString())


class PopupLineEditorParameterItem(ptypes.WidgetParameterItem):
    def __init__(self, param, depth):
        strings = param.opts.get("limits", [])
        self.model = QtCore.QStringListModel(strings)
        param.sigLimitsChanged.connect(
            lambda _param, limits: self.model.setStringList(limits)
        )
        super().__init__(param, depth)

    def makeWidget(self):
        opts = self.param.opts
        editor = PopupLineEditor(
            model=self.model,
            clearOnComplete=False,
            forceMatch=opts.get("forceMatch", True),
            validateCase=opts.get("validateCase", False),
        )
        editor.setValue = editor.setText
        editor.value = editor.text
        editor.sigChanged = editor.editingFinished
        return editor

    def widgetEventFilter(self, obj, ev):
        # Prevent tab from leaving widget
        return False


class PopupLineEditorParameter(Parameter):
    itemClass = PopupLineEditorParameterItem

    def __init__(self, **opts):
        if "value" in opts:
            limits = opts.get("limits", [])
            if opts["value"] not in limits:
                limits.insert(0, opts["value"])
            opts["limits"] = limits
        super().__init__(**opts)


class DisplayValueParameterItem(ptypes.StrParameterItem):
    def makeWidget(self):
        widget = super().makeWidget()
        widget.setValue = lambda val: widget.setText(str(val))
        widget.value = self.param.value
        return widget

    def __init__(self, param, depth):
        super().__init__(param, depth)
        self.widget.setReadOnly(True)
        if param.value() is None:
            # Force set initial value if "None" otherwise it won't get set
            # Because pg.eq() will say value never needed to be initially set
            self.widget.setText("None")
            self.updateDisplayLabel()


class DisplayValueParameter(ptypes.SimpleParameter):
    itemClass = DisplayValueParameterItem

    def __init__(self, **opts):
        opts.update(enabled=False)
        super().__init__(**opts)

    def _interpretValue(self, v):
        return v


ptypes.registerParameterType("display", DisplayValueParameter, override=True)
ptypes.registerParameterItemType("keysequence", KeySequenceParameterItem, override=True)
ptypes.registerParameterType("popuplineeditor", PopupLineEditorParameter, override=True)



from __future__ import annotations

import copy
import functools
import typing as t
from collections.abc import MutableMapping
from contextlib import ExitStack
from pathlib import Path

from pyqtgraph.parametertree import InteractiveFunction, Parameter
from pyqtgraph.parametertree.interactive import Interactor
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets

from qtextras import fns
from qtextras._funcparse import QtExtrasInteractor, RunOptions
from qtextras.params.optionsdict import OptionsDict
from qtextras.typeoverloads import FilePath
from qtextras.widgets.easywidget import EasyWidget

__all__ = [
    "RunOptions",
    "ParameterEditor",
    "ParameterContainer",
]

Signal = QtCore.Signal
_ParentType = t.Union[str, t.Sequence[str], Parameter, None]


class ParameterContainer(MutableMapping):
    """
    Utility for exposing dict-like behavior to the user for inserting and querying
    parameter objects. Behaves amost like a pyqtgraph parameter, with the exception
    that new parameter "children" can be added through the dict interface
    """

    def __init__(self, parameters: dict = None, extra: dict = None, **kwargs):
        if parameters is None:
            parameters = {}
        if extra is None:
            extra = {}
        self.parameters = parameters
        self.extra = extra
        for name, value in kwargs.items():
            self[name] = value

    def __delitem__(self, key):
        if key in self.extra and key in self.parameters:
            raise KeyError(
                f"Key `{key}` is ambiguous, exists in both `extras` and `parameters`"
            )
        try:
            self.extra.__delitem__(key)
        except KeyError:
            # KeyError on non-existence will be raised in the except block
            self.parameters.__delitem__(key)

    def __iter__(self):
        for d in self.extra, self.parameters:
            yield from d

    def __len__(self):
        return self.parameters.__len__() + self.extra.__len__()

    def __setitem__(self, key, value):
        if isinstance(value, Parameter):
            self.parameters[key] = value
            self.extra.pop(key, None)
        elif key in self.parameters:
            self.parameters[key].setValue(value)
            # Assume that if the key is already in parameters, it's not in extras
            # self.extra.pop(key, None)
        else:
            self.extra[key] = value
            self.parameters.pop(key, None)

    def __getitem__(self, item):
        try:
            return self.extra[item]
        except KeyError:
            return self.parameters[item].value()

    def copy(self, deepCopyParams=True, deepCopyExtras=False, deepCopyAttrs=False):
        ret = type(self)()
        ret.__dict__.update(self.__dict__)
        # Will be filled in later
        ret.extra = {}
        ret.parameters = {}
        if deepCopyAttrs:
            ret.__dict__ = copy.deepcopy(ret.__dict__)

        if deepCopyExtras:
            ret.extra = copy.deepcopy(self.extra)
        else:
            ret.extra = self.extra.copy()
        # As of now, ret.copy()['param'] = value will still modify ret['param'] if
        # `deepCopyParams` is false. Sometimes this may be desired, but doesn't follow
        # the interface of dict.copy()['key'] = value preventing modification of the
        # original dict's key
        if deepCopyParams:
            for param, val in self.parameters.items():
                ret.parameters[param] = Parameter.create(**val.saveState())
        else:
            ret.parameters = self.parameters.copy()
        return ret

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memodict):
        return self.copy(deepCopyExtras=True, deepCopyAttrs=True)

    def __str__(self):
        pieces = []
        for subdict, title in zip([self.extra, self.parameters], ["Extra", "Params"]):
            if len(subdict):
                pieces.append(f"{title}: {subdict}")
            else:
                pieces.append(f"{title}: <No keys>")
        return "\n".join(pieces)

    def __repr__(self):
        return str(self)


class ParameterStateSignals(QtCore.QObject):
    # List of string names of new or removed states
    created = Signal(list)
    deleted = Signal(list)
    loadRequested = Signal(str)


class FileStateManager:
    DEFAULT_STATE_NAME = "Default"

    def __init__(
        self,
        directory: FilePath | None = None,
        suffix: str = ".param",
    ):
        if directory:
            directory = Path(directory)

        self.signals = ParameterStateSignals()
        self.directory = directory
        self.suffix = suffix

        self.watcher: QtCore.QFileSystemWatcher | None = None
        self.savedStates: list[str] = []
        """Records which states are available since the last change"""

        if self.directory is not None:
            self.watcher = QtCore.QFileSystemWatcher()
            self.watcher.addPath(str(self.directory.resolve()))
            # False positive type hint issues, return values are allowed
            self.watcher.directoryChanged.connect(self.updateSavedStates)  # noqa
            self.watcher.fileChanged.connect(self._onWatcherFileChanged)
            self.updateSavedStates()

    def updateSavedStates(self):
        """
        Evaluates this directory's saved states and exposes them to the application
        via signals indicating new and deleted files
        """
        dirStates = sorted(f.stem for f in self.directory.glob(f"*{self.suffix}"))

        curStatesSet = set(dirStates)
        savedStatesSet = set(self.savedStates)

        newStates = sorted(curStatesSet - savedStatesSet)
        delStates = sorted(savedStatesSet - curStatesSet)

        self.savedStates = dirStates

        # Only emit if there are states in each category
        newStates and self.signals.created.emit(newStates)
        delStates and self.signals.deleted.emit(delStates)
        return dict(created=newStates, deleted=delStates)

    def _onWatcherFileChanged(self, file: str):
        """
        Called when a watched state file is changed.
        """
        if self.formatFileName(file).exists():
            self.signals.loadRequested.emit(self.formatFileName(file).stem)

    def getDefaultStateName(self, returnStateDict=False) -> Path | tuple[Path, dict]:
        name = self.formatFileName(self.DEFAULT_STATE_NAME)
        if returnStateDict:
            return name, fns.attemptFileLoad(name, missingOk=True)
        return name

    def formatFileName(self, stateName: str | Path = None) -> Path | None:
        if stateName is None:
            return None
        stateName = (self.directory or Path()).joinpath(stateName).resolve()
        if not stateName.suffix:
            stateName = stateName.with_suffix(self.suffix)

        return stateName

    def delete(self, stateName: str):
        filename = self.formatFileName(stateName)
        if not filename.exists():
            return
        filename.unlink()
        # File watcher will take care of emitting deleted signal

    def saveState(self, stateName: str, stateDict: dict):
        if self.directory is None:
            return
        stateName = self.formatFileName(stateName)
        if not self.directory.exists() and stateName.parent == self.directory:
            raise ValueError(
                f"Attempting to save `{stateName}` when `self.directory` does not exist"
            )
        fns.saveToFile(stateDict, stateName)

    def loadState(
        self,
        stateName: str = None,
        stateDict: dict = None,
        addDefaults=False,
        watch=False,
    ):
        filename = self.formatFileName(stateName)
        if stateDict is None:
            if filename is None:
                raise ValueError("Must provide either a filename or a state dict")
            stateDict = fns.attemptFileLoad(filename)
        else:
            # If loading from a dict, we don't need to watch the file
            watch = False
        if addDefaults:
            stateDict = self.maybePopulateDefaults(stateDict)
        if watch and self.watcher:
            if files := self.watcher.files():
                self.watcher.removePaths(files)
            self.watcher.addPath(str(self.directory))
            self.watcher.addPath(str(filename))
        return stateDict

    def maybePopulateDefaults(self, stateDict: dict):
        """
        Add default parameters to the state dict via hierarchical updates. This
        is only performed if there is a default state ("DEFAULT_STATE_NAME" in
        ``self.directory``) that can be loaded
        """
        name, defaultState = self.getDefaultStateName(returnStateDict=True)
        if defaultState:
            stateDict = fns.hierarchicalUpdate(defaultState, stateDict)
        return stateDict

    def addSavedStatesToMenu(
        self, parentMenu: QtWidgets.QMenu, removeExistingChildren=True, keepUpdated=True
    ):
        """
        Helper function for populating menu from directory contents
        """
        # We don't want all menu children to be removed, since this would also remove
        # the 'edit' and separator options. So, do this step manually. Remove all
        # actions after the separator
        if self.directory is None:
            return
        dirGlob = fns.naturalSorted(
            f.stem for f in self.directory.glob(f"*{self.suffix}")
        )

        if removeExistingChildren:
            parentMenu.clear()

        # TODO: At the moment parameter files that start with '.' aren't getting included
        #  in the glob

        def loaderFunc(stateName):
            self.signals.loadRequested.emit(stateName)

        def populateFunc():
            self.addSavedStatesToMenu(parentMenu)

        for name in dirGlob:
            # glob returns entire filepath, so keep only filename as layout name
            curAction = parentMenu.addAction(name)
            curAction.triggered.connect(functools.partial(loaderFunc, name))

        if keepUpdated:
            self.signals.created.connect(populateFunc)
            self.signals.deleted.connect(populateFunc)


class ParameterEditor(QtWidgets.QWidget):
    sigStateCreated = Signal(str)
    sigStateChanged = Signal(object)
    # Dict of changes, or *None*
    sigStateDeleted = Signal(str)

    sigFunctionRegistered = Signal(object)

    name: str = None
    """Human readable name (i.e., used when inserting into a menu)"""

    defaultInteractor = QtExtrasInteractor()
    defaultParent: _ParentType = None

    def __init__(
        self, *, name: str = None, directory: FilePath = None, suffix=".param"
    ):
        """
        GUI controls for user-interactive parameters within a QtWidget (usually main
        window). Each window consists of a parameter tree and basic saving capabilities.

        Parameters
        ----------
        name
            User-readable name of the parameter editor. If *None*, :attr:`name`
            is used
        directory
            Directory to save parameter files to. If *None*, no saving is performed
        suffix
            The suffix of the saved settings. E.g. if a settings configuration is saved
            with the name "test", it will result in a file "test[suffix]"
        """
        super().__init__()
        self.stateManager = mgr = FileStateManager(directory, suffix)

        self.stateName = self.getDefaultStateName().stem

        self.name = name

        self.nameFunctionMap: dict[str, InteractiveFunction] = {}
        """Record of registered functions / processes, {function.__name__: function}"""
        self.nameParameterMap: dict[str, Parameter] = {}
        """Record of registered parameters, {parameter.name(): parameter}"""
        self.nameShortcutMap: dict[str, QtGui.QShortcut | t.Any] = {}
        """Record of InteractiveFunction shortcuts, {function.__name__: shortcut}"""

        # -----------
        # Construct parameter tree
        # -----------
        self.rootParameter = Parameter.create(name="Parameters", type="group")
        self.tree = fns.flexibleParameterTree(self.rootParameter, showTop=False)

        self._buildGui()

        # Connect last on the off chance that files are being manipulated during
        # the construction of the GUI
        mgr.signals.loadRequested.connect(self.onLoadRequested)

    @property
    def directory(self):
        return self.stateManager.directory

    @property
    def suffix(self):
        return self.stateManager.suffix

    def getDefaultStateName(self, returnStateDict=False):
        return self.stateManager.getDefaultStateName(returnStateDict)

    def formatFileName(self, stateName: str):
        return self.stateManager.formatFileName(stateName)

    def saveParameterValues(
        self,
        stateName: str = None,
        stateDict: dict = None,
        *,
        addDefaults=False,
        blockWrite=False,
        requestLoad: bool | None = None,
    ) -> t.Dict[str, t.Any]:
        """
        Saves the current parameter values to a file.

        Parameters
        ----------
        stateName
            Name of the state to save. If *None*, the current state name is used.
        stateDict
            Dictionary of parameter values to save. If *None*, the current parameter
            values are used.
        addDefaults
            If *True*, include default values in the saved state. These can either
            be the default values of parameters in the tree, or the default state
            if one exists in the directory.
        blockWrite
            If *True*, don't actually write the file. This is useful for capturing
            the saved state of a parameter editor without writing to disk.
        requestLoad
            If *True*, emit the :attr:`stateManager.signals.loadRequested` signal. If
            *None*, this is determined by whether the state name has changed. Useful
            for making sure that the state name and values are in sync when i.e. the
            user has created a new state.
        """
        if stateName is None:
            stateName = self.stateName
        if stateName is None:
            raise ValueError(
                "No state name given and no default state name set; cannot save"
            )
        if stateDict is None:
            stateDict = self.getParameterValues()
        if addDefaults:
            _, defaults = self.stateManager.getDefaultStateName(returnStateDict=True)
            if defaults is None:
                defaults = self.getParameterDefaults()
            stateDict = fns.hierarchicalUpdate(defaults, stateDict)
        if not blockWrite:
            # Don't trigger infinite "load" loops through file watcher signal
            # if saving self's current state
            self._saveAndMaybeBlockUpdate(stateName, stateDict)
        if requestLoad is None:
            fmt = self.formatFileName
            requestLoad = fmt(stateName) != fmt(self.stateName)
        if requestLoad and not blockWrite:
            self.stateManager.signals.loadRequested.emit(stateName)
        return stateDict

    def getParameterDefaults(self):
        return fns.parameterValues(self.rootParameter, value="default")

    def getParameterValues(self):
        return fns.parameterValues(
            self.rootParameter, valueFilter=fns.valueIsNotDefault
        )

    def _saveAndMaybeBlockUpdate(self, stateName: str, stateDict: dict):
        """
        Save state and block updates to prevent infinite loop when saving self's
        current state
        """
        mgr = self.stateManager
        fmt = mgr.formatFileName
        with ExitStack() as stack:
            if fmt(stateName) == fmt(self.stateName):
                stack.enter_context(fns.makeDummySignal(mgr.signals, "loadRequested"))
            mgr.saveState(stateName, stateDict)

    def saveParameterValuesGui(self):
        saveName = fns.dialogGetSaveFileName(self, "Save As", self.stateName)
        if saveName == (default := self.getDefaultStateName().stem):
            QtWidgets.QMessageBox.information(
                self,
                "Save As",
                f"`{default}` is automatically generated and will overwrite changes "
                f"on next startup. Please choose a different name.",
            )
            return self.saveParameterValuesGui()
        if saveName:
            return self.saveParameterValues(saveName)

    def _parseAndUpdateCandidateParameters(
        self, loadDict: dict, candidateParameters: t.Sequence[Parameter] = None
    ):
        if candidateParameters is None:
            candidateParameters = fns.flattenedParameters(
                self.rootParameter, groupTypes=("group", "_actiongroup")
            )
        if not candidateParameters:
            return

        def validName(param, name):
            return name in (param.opts["title"], param.name())

        def checkParentChain(param, name):
            if not param:
                return False
            return validName(param, name) or checkParentChain(param.parent(), name)

        unhandled = {}
        # Copy for mutable object
        for kk, vv in loadDict.items():
            if isinstance(vv, dict):
                # Successively traverse down child tree
                curCandidates = [
                    p for p in candidateParameters if checkParentChain(p, kk)
                ]
                # By this point defaults are already populated so no need to reset them
                # otherwise infinite recursion will occur
                self._parseAndUpdateCandidateParameters(
                    vv, candidateParameters=curCandidates
                )
            else:
                unhandled[kk] = vv
        with self.rootParameter.treeChangeBlocker():
            for kk, vv in unhandled.items():
                matches = [p for p in candidateParameters if validName(p, kk)]
                if len(matches) == 1:
                    matches[0].setValue(vv)
                # elif len(matches) == 0:
                #   warnings.warn(
                #       f'No matching parameters for key {kk}. Ignoring.', UserWarning
                #   )
                elif len(matches) > 0:
                    raise ValueError(
                        f"Multiple matching parameters for key {kk}:\n" f"{matches}"
                    )

    def loadParameterValues(
        self,
        stateName: str | Path = None,
        stateDict: dict = None,
        addDefaults=False,
        candidateParameters: t.List[Parameter] = None,
    ) -> dict[str, t.Any]:
        """
        Can restore a state created by ``fns.parameterValues``

        Parameters
        ----------
        stateName
            Name of the state to load
        stateDict
            If given, this dictionary will be used instead of loading from the state
            manager
        addDefaults
            If *True*, the default values of the parameters will be used if the
            parameter is not found in the state dictionary. If a state with name
            ``self.DEFAULT_STATE_NAME`` is present in ``self.directory``, this will
            be used as the default state. Otherwise, the default values of the
            parameters will be used.
        candidateParameters
            If given, only these parameters will be considered for loading. This is
            useful if you want to load a subset of parameters from a state.

        Returns
        -------
        Loaded state dictionary. If ``stateDict`` was given, this will be the same
        object unless ``addDefaults`` is *True*.
        """
        if candidateParameters is not None and not len(candidateParameters):
            # Nothing to update, no reason to load state dict
            self.stateName = self.stateManager.formatFileName(stateName).stem
            return self.getParameterValues()

        stateDict = self.stateManager.loadState(
            stateName, stateDict, addDefaults, watch=True
        )
        if addDefaults and not self.stateManager.getDefaultStateName().exists():
            # Defaults need to be added
            stateDict = fns.hierarchicalUpdate(self.getParameterDefaults(), stateDict)

        self._parseAndUpdateCandidateParameters(stateDict, candidateParameters)
        self.stateName = self.stateManager.formatFileName(stateName).stem

        return self.getParameterValues()

    def onLoadRequested(self, stateName):
        # First check for extra keys, will be the case if 'children' is one of the
        # keys. Can't do value-loading in that case, must do state-loading instead
        stateDict = self.stateManager.loadState(stateName)
        if not stateDict:
            # File didn't exist or there was a problem during loading
            return
        if "children" in stateDict:
            # Bug in pyqtgraph restore state doesn't play nice when parameters have
            # connected functions outside the parameter item, so re-implement without
            # inserting or removing children
            with self.rootParameter.treeChangeBlocker():
                fns.applyParameterOpts(self.rootParameter, stateDict)
        else:
            self.loadParameterValues(stateName, stateDict)

    def _resolveParameterParent(self, parent: _ParentType = None) -> Parameter:
        parent = parent or self.defaultParent
        if parent is None:
            parent = self.rootParameter
        if isinstance(parent, Parameter):
            return parent
        if isinstance(parent, str):
            parent = [parent]
        return fns.getParameterChild(self.rootParameter, *parent)

    def registerParameterList(
        self, parameters: t.Sequence[OptionsDict | Parameter], *args, **kwargs
    ) -> list[Parameter]:
        """
        Registers a list of properties and returns an array of each. For parameter
        descriptions, see :meth:`regiseterParameter`.
        """
        outProps = []
        with self.rootParameter.treeChangeBlocker():
            for param in parameters:
                outProps.append(self.registerParameter(param, *args, **kwargs))
        return outProps

    def registerParameter(
        self,
        parameter: OptionsDict | Parameter,
        *,
        parent: _ParentType = None,
        container: ParameterContainer = None,
        **extraOpts,
    ) -> Parameter:
        """
        Registers a property defined by ``constParam`` that will appear in the
        respective parameter editor.

        Parameters
        ----------
        parameter
            The parameter to register. Can either be an instantiated Parmeter
            or ``OptionsDict`` specification.
        parent
            See ``parent`` documentation in :meth:`ParameterEditor.registerFunction`
        container
            Optional container in which to insert this parameter
        **extraOpts
            Extra options passed directly to the created Parameter

        Returns
        -------
        Property bound to this value in the parameter editor
        """
        if isinstance(parameter, Parameter):
            paramOpts = parameter.opts
        else:
            paramOpts = dict(parameter)
        paramOpts.update(extraOpts)

        parent = self._resolveParameterParent(parent)
        paramForEditor = fns.getParameterChild(parent, childOpts=paramOpts)

        if container is not None:
            container[parameter] = paramForEditor
        return paramForEditor

    def registerFunction(
        self,
        function: t.Callable,
        *,
        name: str = None,
        interactor: Interactor = None,
        parent: _ParentType = None,
        runActionTemplate: dict = None,
        parametersNeedRunKwargs: bool | None = True,
        container: ParameterContainer = None,
        **kwargs,
    ) -> InteractiveFunction:
        """
        Like :meth:`registerParameter` but for functions instead along with interactive
        parameters for each argument. A button is added for the user to force run this
        function as well. In the case of a function with no parameters, the button will
        be named the same as the function itself for simplicity.

        Parameters
        ----------
        function
            Function to make interactive
        name:
            Name of the function. If None, defaults to the ``function.__name__``
        interactor
            The interactor to use for creating Parameters out of each argument. If None,
            defaults to :attr:`self.defaultInteractor`.
        parent
            Top :class:`Parameter` for registration. If a string or tuple is provided,
            it will be converted to a Parameter using ``fns.getParameterChild`` and
            inserted as a descendant of :attr:`self.rootParameter`. If None, defaults
            to :attr:`self.rootParameter`.
        runActionTemplate
            Template for the run action. See ``Interactor.runActionTemplate`` for
            more details.
        parametersNeedRunKwargs
            Passed to ``InteractiveFunction.parametersNeedRunKwargs``. If ``None``,
            the function's original attribute value is unchanged.
        container
            The container in which to put child parameters from the registered function.
            This has a similar effect to calling `registerProps` on each of the
            function arguments with this container. *Note*: while the former will
            register props according to the `parameter` OptionsDict, this will register
            to the container using the raw parameter name. A KeyError is raised when
            a new parameter name clashes with something already in the container.
        **kwargs
            All additional kwargs are passed to InteractiveFunction when wrapping the
            function. If an InteractiveFunction is passed in, ``kwargs`` are passed to
            the interactor.
        """
        if interactor is None:
            interactor = self.defaultInteractor
        rawFuncGiven = not isinstance(function, InteractiveFunction)
        if rawFuncGiven:
            function = InteractiveFunction(function)
        runActionTemplate = dict(runActionTemplate) if runActionTemplate else {}

        if rawFuncGiven and name is None and "name" in runActionTemplate:
            name = runActionTemplate.get("name")
            defaultName = self.defaultInteractor.runActionTemplate["defaultName"]
            runActionTemplate["name"] = defaultName
        elif not rawFuncGiven is not None and rawFuncGiven:
            raise ValueError(
                "Cannot specify `name` override when passing an already-interactive "
                "function"
            )

        parent = self._resolveParameterParent(parent)

        if name is not None:
            function.__name__ = name
        if parametersNeedRunKwargs is not None:
            function.parametersNeedRunKwargs = parametersNeedRunKwargs
        if runActionTemplate.get("shortcut"):
            self.registerFunctionShortcut(function, **runActionTemplate)
            del runActionTemplate["shortcut"]
        self.nameFunctionMap[function.__name__] = function
        if runActionTemplate:
            kwargs["runActionTemplate"] = runActionTemplate

        self._interactAndHandleExistingParameters(
            interactor, function, parent=parent, **kwargs
        )

        if container is None:
            self.sigFunctionRegistered.emit(function)
            return function

        funcKwargs = {**function.extra, **function.parameters}
        if alreadyExists := set(container) & set(funcKwargs):
            raise KeyError(f"{alreadyExists} already exist in {container}")

        for name, value in funcKwargs.items():
            container[name] = value
        self.sigFunctionRegistered.emit(function)
        return function

    def registerFunctionShortcut(
        self,
        function: t.Callable,
        shortcut: str,
        ownerWidget: QtWidgets.QWidget = None,
        force=False,
        **opts,
    ):
        name = function.__name__
        if name in self.nameShortcutMap and not force:
            raise KeyError(
                f"{name} already present in shortcut map: {self.nameShortcutMap[name]=}"
            )
        if ownerWidget is None:
            defaultContext = QtCore.Qt.ShortcutContext.ApplicationShortcut
            ownerWidget = self
        else:
            defaultContext = QtCore.Qt.ShortcutContext.WidgetWithChildrenShortcut
        context = opts.get("context", defaultContext)
        qShortcut = QtGui.QShortcut(shortcut, ownerWidget, context=context)
        qShortcut.activated.connect(function)
        self.nameShortcutMap[name] = qShortcut
        return qShortcut

    def registerObjectShortcut(
        self, shortcuttable: t.Any, shortcut: str, name: str, force=False, **opts
    ):
        """
        Objects like ``QAction`` and ``QPushButton`` can be given shortcuts, but don't
        give access to a ``QShortcut`` that is directly manipulatable. Instead,
        their shortcut is given and accessed via its key sequence. Play nicely with
        these objects, but otherwise behave similarly to
        :meth:`ParameterEditor.registerFunctionShortcut`
        """
        if name in self.nameShortcutMap and not force:
            raise KeyError(
                f"{name} already present in shortcut map: {self.nameShortcutMap[name]=}"
            )
        if not hasattr(shortcuttable, "setShortcut"):
            raise ValueError(f"{shortcuttable} must have `setShortcut()` method")
        shortcuttable.setShortcut(shortcut)
        self.nameShortcutMap[name] = shortcuttable

    @classmethod
    def _interactAndHandleExistingParameters(cls, interactor, function, **kwargs):
        existingParams = function.parameters.copy()
        # Set to nothing before interacting, so they can be added back without
        # being disconnected
        function.parameters = {}
        if alreadyRegistered := set(existingParams).intersection(kwargs):
            raise ValueError(
                f"Cannot pass parameters {alreadyRegistered} that are already "
                f"registered to {function}"
            )
        kwargs["ignores"] = set(kwargs.get("ignores", ())) | set(existingParams)
        interactor(function, **kwargs)
        function.parameters.update(existingParams)
        for ii, (name, param) in enumerate(existingParams.items()):
            kwargs["parent"].insertChild(ii, param)
            del function.extra[name]
            function.parameterCache[name] = param.value()

    def createWindowDock(
        self,
        window: QtWidgets.QMainWindow,
        title: str = None,
        dockArea=QtCore.Qt.DockWidgetArea.RightDockWidgetArea,
        createProcessMenu=True,
        addShowAction=True,
        menu: QtWidgets.QMenu = None,
    ) -> tuple[QtWidgets.QDockWidget, QtWidgets.QMenu | None]:
        """
        Creates a dock widget for this parameter editor and adds it to the given
        window. If no window is provided, a new one is created

        Parameters
        ----------
        window
            Window to add the dock to
        title
            Title for the dock widget and created menu
        dockArea
            Area of the window to dock to
        createProcessMenu
            Whether to create a menu for running registered editor functions.
            If ``True``, a menu is created with the name ``title`` and added
            to the window's menu bar.
            .. note::
                This will change the return value to a tuple of the dock and the menu.
        addShowAction
            Whether to add an action to the created process menu which
            toggles display of the dock widget. Note that if ``createProcessesMenu``
            is ``False``, this is ignored

        Returns
        -------
        dock or tuple of (dock, menu)
            Either the dock widget or both the dock and created menu, the latter when
            ``createProcessesMenu`` is ``True``
        """
        title = title or self.name.title()
        dock = QtWidgets.QDockWidget(title)
        dock.setObjectName(title)
        dock.setWidget(self)
        dock.hide()

        curAreaWidgets = [
            d
            for d in window.findChildren(QtWidgets.QDockWidget)
            if window.dockWidgetArea(d) == dockArea
        ]
        window.addDockWidget(dockArea, dock)
        if len(curAreaWidgets):
            window.tabifyDockWidget(curAreaWidgets[-1], dock)
            window.setTabPosition(dockArea, QtWidgets.QTabWidget.TabPosition.North)

        if not createProcessMenu and not addShowAction:
            return dock, None

        def onShow():
            dock.show()
            dock.raise_()

        if menu is None:
            menu = QtWidgets.QMenu(title)

        if addShowAction:
            menu.addAction(f"{self.name}...", onShow)
        if createProcessMenu:
            self.createActionsFromFunctions(menu)
        return dock, menu

    def createActionsFromFunctions(
        self,
        menu: QtWidgets.QMenu = None,
        stealShortcuts=True,
        keepUpdated=True,
    ) -> QtWidgets.QMenu:
        if menu is None:
            menu = QtWidgets.QMenu(fns.nameFormatter(self.name))

        def addProcess(process):
            name = process.__name__
            if name not in self.rootParameter.names:
                raise ValueError(
                    "Creating a process menu from children outside `self.rootParameter` "
                    "is currently not supported."
                )
            parameter = self.rootParameter.child(name)
            # Don't interact with functions unless they specify RunOptions.ON_ACTION
            if not parameter.opts.get("button", {}).get("visible", True):
                return
            actTitle = self.defaultInteractor.titleFormat(process.__name__)
            act = menu.addAction(actTitle, process.__call__)
            if stealShortcuts and (
                shortcut := self.nameShortcutMap.get(process.__name__)
            ):
                self.registerObjectShortcut(
                    act, shortcut.key(), process.__name__, force=True
                )

        for proc in self.nameFunctionMap.values():
            addProcess(proc)
        if keepUpdated:
            self.sigFunctionRegistered.connect(addProcess)
        return menu

    # Allow customization through kwargs in subclasses
    # noinspection PyUnusedLocal
    def _buildGui(self, **kwargs):
        self.setWindowTitle(self.name)
        self.setObjectName(self.name)

        # -----------
        # Additional widget buttons
        # -----------
        self.expandAllButton = QtWidgets.QPushButton("Expand All")
        self.collapseAllButton = QtWidgets.QPushButton("Collapse All")
        self.saveAsButton = QtWidgets.QPushButton("Save As...")
        pol = QtWidgets.QSizePolicy.Policy
        self.saveAsButton.setSizePolicy(pol.MinimumExpanding, pol.Fixed)
        self._createLoadButton()

        # -----------
        # Widget layout
        # -----------
        # Optional buttons that can be specified in subclasses under `_guiChildren()`
        # if they should appear
        self.treeButtonsWidget = EasyWidget(
            [self.expandAllButton, self.collapseAllButton]
        )
        self.treeButtonsWidget.widget_.hide()
        children = self._guiChildren()

        EasyWidget.buildWidget(children, baseWidget=self)

        if self.stateManager.directory is None:
            self.saveAsButton.hide()
            self.loadButton.hide()

        self.tree.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        # -----------
        # UI Element Signals
        # -----------
        self.expandAllButton.clicked.connect(
            lambda: fns.setParametersExpanded(self.tree)
        )
        self.collapseAllButton.clicked.connect(
            lambda: fns.setParametersExpanded(self.tree, False)
        )
        self.saveAsButton.clicked.connect(self.saveParameterValuesGui)

    def _guiChildren(self) -> list:
        """
        Specifies which children should be added to this GUI. The return value
        is directly passed to :func:`EasyWidget.buildWidget`
        """
        return [
            self.treeButtonsWidget,
            self.tree,
            [self.saveAsButton, self.loadButton],
        ]

    def _createLoadButton(self):
        self.loadButton = QtWidgets.QToolButton()
        self.loadButton.setText("Load...")
        pol = QtWidgets.QSizePolicy.Policy
        self.loadButton.setSizePolicy(pol.MinimumExpanding, pol.Fixed)
        self.loadButton.setMenu(QtWidgets.QMenu(self.loadButton))
        self.loadButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.stateManager.addSavedStatesToMenu(self.loadButton.menu())
        return self.loadButton

    def __repr__(self):
        selfCls = type(self)
        oldName: str = super().__repr__()
        # Remove module name for brevity
        oldName = oldName.replace(
            f"{selfCls.__module__}.{selfCls.__name__}",
            f"{selfCls.__name__} '{self.name}'",
        )
        return oldName



from __future__ import annotations

import builtins
from typing import Optional, Sequence

import numpy as np
import pandas as pd


def _rescale(data, min_, max_):
    # Handle nan values
    rng = max_ - min_
    if rng == 0:
        return np.ones_like(data)
    return (data.astype(float) - min_) / rng


class _UNSPECIFIED_DEFAULT:
    pass


class OptionsDict:
    comparator = str
    """
    Function for determining equality when comparing to another object
    """

    def __init__(
        self, name: str, value=None, type: Optional[str] = None, tip="", **opts
    ):
        """
        Encapsulates parameter information use within the main application

        Parameters
        ----------
        name
            Display name of the parameter
        value
            Initial value of the parameter. This is used within the program to infer
            parameter type, shape, comparison methods, etc.
        type
            Type of the variable if not easily inferrable from the value itself. For
            instance, class:`ShortcutParameter<shortcuts.ShortcutParameter>` is
            indicated with string values (e.g. 'Ctrl+D'), so the user must explicitly
            specify that such an :class:`OptionsDict` is of type 'shortcut' (as defined
            in :class:`ShortcutParameter<pgregistered.shortcut.ShortcutParameter>`) If
            the type *is* easily inferrable, this may be left blank.
        tip
            Additional documentation for this parameter.
        **opts
            Additional options associated with this parameter
        """
        if opts is None:
            opts = {}
        if type is None:
            # Infer from value
            type = builtins.type(value).__name__.lower()
        self.name = name
        self.value = value
        self.type = type
        self.tip = tip or opts.get("tip", "")

        self.opts = opts

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return f"{self.name}: <{self.type}>"

    def __lt__(self, other):
        """
        Required for sorting by value in component table. Defer to alphabetic
        sorting

        Parameters
        ----------
        other
            Other :class:`OptionsDict` member for comparison
        """
        return self.comparator(self) < self.comparator(other)

    def __eq__(self, other):
        # TODO: Highly naive implementation. Be sure to make this more robust if
        #  it needs to be for now assume only other prjparamss will be passed in
        return self.comparator(self) == self.comparator(other)

    def __hash__(self):
        # Since every parameter within a group will have a unique name, just the name is
        # sufficient to form a proper hash
        return hash(
            self.comparator(self),
        )

    def toNumeric(self, data: Sequence, rescale=False, returnMapping=False):
        """
        Useful for converting string-like or list-like parameters to integer
        representations.

        If self's `value` is non-numeric data (e.g. strings):
           - First, the parameter is searched for a 'limits' property. This will
             contain all possible values this field could be.
           - If no limits exist, unique values in the input data will be considered
             the limits

        Parameters
        ----------
        data
            Array-like data to be turned into numeric values according to the parameter
            type
        rescale
            If *True*, values will be rescaled to the range 0-1
        returnMapping
            Whether to also return a series of the unique values within ``data``
            in the form {numericVal: label}.
        """
        numericVals = np.asarray(data).copy()
        if len(data) == 0:
            mapping = pd.Series(name=self, dtype=object)
            if returnMapping:
                return numericVals, mapping
            return numericVals
        if not np.issubdtype(type(self.value), np.number):
            # Check for limits that indicate the exhaustive list of possible values.
            # Otherwise, just use the unique values of this set as the limits
            if "limits" in self.opts:
                listLims = list(self.opts["limits"])
                numericVals = np.array([listLims.index(v) for v in numericVals])
                mapping = pd.Series(listLims, np.arange(len(listLims), dtype=int))
            else:
                listLims, numericIdxs, numericVals = np.unique(
                    numericVals, return_index=True, return_inverse=True
                )
                mapping = pd.Series(listLims, numericVals[numericIdxs], name=self)
        elif returnMapping:
            # Potentially expensive, only compute if requested
            uniques = np.unique(numericVals)
            mapping = pd.Series(uniques, uniques, name=self)
        else:
            # Dummy mapping to allow ops
            class mapping:
                index = np.array([numericVals.max(), numericVals.min()])

        if rescale:
            min_, max_ = mapping.index.min(), mapping.index.max()
            # Make sure to reduce min to account for offset in reverse direction
            numericVals = _rescale(numericVals, min_, max_)
            mapping.index = _rescale(mapping.index, min_, max_)
        if returnMapping:
            return numericVals, mapping
        return numericVals

    def __getitem__(self, item):
        if item in self._fields:
            return getattr(self, item)
        return self.opts[item]

    def __setitem__(self, key, value):
        if key in self._fields:
            setattr(self, key, value)
        else:
            self.opts[key] = value

    def __contains__(self, item):
        return item in [*self._fields, *self.opts.keys()]

    def __delitem__(self, key):
        del self.opts[key]

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            if default is _UNSPECIFIED_DEFAULT:
                raise
            return default

    @property
    def _fields(self):
        excludeFields = {"opts"}
        return [field for field in self.__dict__ if field not in excludeFields]

    def keys(self):
        return self._fields + list(self.opts.keys())

    def __contains__(self, item):
        return item in self._fields or item in self.opts

    def addHelpText(self, prependText="", postfixText="", replace=True, newline="<br>"):
        helpText = self.tip
        if len(prependText) > 0 and len(helpText) > 0 or len(postfixText) > 0:
            prependText += newline
        curText = prependText + self.tip
        if len(postfixText) > 0:
            curText += newline + postfixText
        if replace:
            self.tip = curText
        return curText



from __future__ import annotations

import inspect
import typing as t

from pyqtgraph import QtGui, QtWidgets
from pyqtgraph.parametertree import Parameter

from qtextras.fns import flattenedParameters, forceRichText
from qtextras.params import OptionsDict

ButtonCallable = t.Union[t.Callable[[OptionsDict], t.Any], t.Callable[[], t.Any]]


class ButtonCollection(QtWidgets.QGroupBox):
    def __init__(
        self,
        parent=None,
        title: str = None,
        buttonParameters: t.Collection[OptionsDict] = (),
        triggerFunctions: t.Union[ButtonCallable, t.Collection[ButtonCallable]] = (),
        exclusive=True,
        asToolButton=True,
        **createOpts,
    ):
        super().__init__(parent)
        self.lastTriggered: t.Optional[OptionsDict] = None
        self.uiLayout = QtWidgets.QHBoxLayout(self)
        self.buttonGroup = QtWidgets.QButtonGroup(self)
        self.optionsFunctionMap: t.Dict[OptionsDict, ButtonCallable] = dict()
        self.optionsButtonMap: t.Dict[OptionsDict, QtWidgets.QPushButton] = dict()
        self.asToolButton = asToolButton
        if title is not None:
            self.setTitle(title)
        self.buttonGroup.setExclusive(exclusive)

        if not isinstance(triggerFunctions, t.Iterable):
            triggerFunctions = [triggerFunctions] * len(buttonParameters)
        for param, fn in zip(buttonParameters, triggerFunctions):
            self.createAndAddButton(param, fn, **createOpts)

    def createAndAddButton(
        self,
        buttonOptions: OptionsDict,
        triggerFunction: ButtonCallable,
        checkable=False,
        **createOpts,
    ) -> QtWidgets.QAbstractButton | None:
        if buttonOptions in self.optionsButtonMap:
            # Either already exists or wasn't designed to be a button
            return None

        if self.asToolButton:
            createOpts.setdefault("buttonType", QtWidgets.QToolButton)
        newBtn = self.createButton(buttonOptions, **createOpts)
        if checkable:
            newBtn.setCheckable(True)
            oldTriggerFn = triggerFunction
            numArgs = self._getTriggerArgumentCount(triggerFunction)

            def newTriggerFn(param: OptionsDict):
                """
                If the button is checkable, only call this function when the button
                is checked
                """
                if newBtn.isChecked():
                    return oldTriggerFn(param) if numArgs else oldTriggerFn()

            triggerFunction = newTriggerFn
        newBtn.clicked.connect(lambda: self.callAssociatedFunction(buttonOptions))

        self.addButton(buttonOptions, newBtn, triggerFunction)
        return newBtn

    def clear(self):
        for button in self.optionsButtonMap.values():
            self.buttonGroup.removeButton(button)
            self.uiLayout.removeWidget(button)
            button.deleteLater()

        self.optionsButtonMap.clear()
        self.optionsFunctionMap.clear()

    def addFromExisting(
        self, other: ButtonCollection, which: t.Collection[OptionsDict] = None
    ):
        for (param, btn), func in zip(
            other.optionsButtonMap.items(), other.optionsFunctionMap.values()
        ):
            if which is None or param in which:
                self.addButton(param, btn, func)

    def addButton(
        self,
        option: OptionsDict,
        button: QtWidgets.QPushButton,
        function: ButtonCallable,
    ):
        self.buttonGroup.addButton(button)
        self.uiLayout.addWidget(button)
        self.optionsFunctionMap[option] = function
        self.optionsButtonMap[option] = button

    @classmethod
    def createButton(
        cls,
        buttonOptions: OptionsDict,
        baseButton: QtWidgets.QAbstractButton = None,
        buttonType=QtWidgets.QPushButton,
        parent=None,
    ):
        tooltipText = buttonOptions.tip
        title = buttonOptions.get("title", buttonOptions.name)
        if baseButton is not None:
            newBtn = baseButton
        else:
            newBtn = buttonType(parent)
            newBtn.setText(title)
        icon = (
            buttonOptions.opts["button"].get("icon")
            if "group" in buttonOptions.type
            else buttonOptions.get("icon")
        )
        if icon:
            newBtn.setText("")
            newBtn.setIcon(QtGui.QIcon(str(icon)))
            tooltipText = buttonOptions.addHelpText(title)
        if tooltipText:
            tooltipText = forceRichText(tooltipText)
            newBtn.setToolTip(tooltipText)
        return newBtn

    def callAssociatedFunction(self, options: OptionsDict):
        if options is None:
            return
        # Ensure function is called in the event it requires a button to be checked
        btn = self.optionsButtonMap[options]
        if btn.isCheckable():
            btn.setChecked(True)
        function = self.optionsFunctionMap[options]
        argCount = self._getTriggerArgumentCount(function)
        ret = function(options) if argCount else function()
        self.lastTriggered = options
        return ret

    def addByParameter(self, parameter: Parameter, **registerOpts):
        """
        Adds a button to a group based on the parameter. Also works for group params
        that have an action nested.
        """
        for parameter in filter(
            lambda p: "action" in p.type(), flattenedParameters(parameter)
        ):
            self.createAndAddButton(
                OptionsDict(**parameter.opts), parameter.activate, **registerOpts
            )

    @classmethod
    def fromToolsEditors(
        cls,
        editors: t.Sequence[t.Any],
        title="Tools",
        ownerCollection: ButtonCollection = None,
        **registerOpts,
    ):
        if ownerCollection is None:
            ownerCollection = cls(title=title, exclusive=True)

        for editor in editors:
            ownerCollection.addByParameter(editor.rootParameter, **registerOpts)

        return ownerCollection

    def toolbarFormat(self):
        """
        Returns a list of buttons + title in a format that's easier to add to a
        toolbar, e.g. doesn't require as much horizontal space
        """
        title = self.title()
        out: t.List[QtWidgets.QWidget] = (
            [] if title is None else [QtWidgets.QLabel(self.title())]
        )
        for btn in self.optionsButtonMap.values():
            out.append(btn)
        return out

    def _getTriggerArgumentCount(self, triggerFunction: t.Callable):
        signature = inspect.signature(triggerFunction)
        numArgs = None
        for bindAttempt in [], [OptionsDict("dummy")]:
            try:
                signature.bind(*bindAttempt)
                numArgs = len(bindAttempt)
            except TypeError:
                pass
        if numArgs is None:
            raise ValueError(
                f"`{triggerFunction}` must accept either 0 or 1 positional arguments"
            )

        return numArgs



from __future__ import annotations

import typing as t

from pyqtgraph import QtCore, QtWidgets
from pyqtgraph.parametertree import Parameter

from qtextras import fns

_layoutTypes = t.Union[t.Literal["H"], t.Literal["V"]]


class EasyWidget:
    def __init__(
        self,
        children: t.MutableSequence,
        layout: str = None,
        useSplitter=False,
        baseWidget: QtWidgets.QWidget = None,
    ):
        if baseWidget is None:
            baseWidget = QtWidgets.QWidget()
        self._built = False
        self.children_ = children
        self.useSplitter = None
        self.widget_ = baseWidget
        self.layout_ = None

        self._resetOpts(useSplitter, layout)

    def _resetOpts(self, useSplitter, layout):
        if layout == "V":
            orient = QtCore.Qt.Orientation.Vertical
            layout = QtWidgets.QVBoxLayout
        elif layout == "H":
            orient = QtCore.Qt.Orientation.Horizontal
            layout = QtWidgets.QHBoxLayout
        else:
            orient = layout = None
        self.orient_ = orient

        if useSplitter == self.useSplitter and self.layout_:
            return
        # Had children in existing widget which will be discarded when changing self
        # widget_ to splitter
        if self.widget_.children() and useSplitter:
            raise ValueError(
                "Cannot change splitter status to *True* when widget already has "
                "children"
            )
        self.useSplitter = useSplitter

        if useSplitter:
            self.layout_ = QtWidgets.QSplitter(orient)
            self.widget_ = self.layout_
        else:
            try:
                self.layout_ = layout()
                self.widget_.setLayout(self.layout_)
            except TypeError:
                # When layout is none
                self.layout_ = None

    def build(self):
        if self._built:
            return
        if self.layout_ is None:
            raise ValueError(
                'Top-level orientation must be set to "V" or "H" before adding children'
            )
        if self.orient_ == QtCore.Qt.Orientation.Horizontal:
            chSuggested = "V"
        elif self.orient_ == QtCore.Qt.Orientation.Vertical:
            chSuggested = "H"
        else:
            chSuggested = None

        for ii, child in enumerate(self.children_):
            morphChild = self.addChild(child, chSuggested)
            if morphChild is not child:
                self.children_[ii] = morphChild
        self._built = True

    def addChild(
        self,
        child: t.Union[QtWidgets.QWidget, t.Sequence, EasyWidget],
        suggestedLayout: str = None,
    ):
        if isinstance(child, QtWidgets.QWidget):
            self.layout_.addWidget(child)
        else:
            child = self.listChildrenWrapper(child, suggestedLayout)
            # At this point, child should be an EasyWidget
            child.build()
            self.layout_.addWidget(child.widget_)
        return child

    def insertChild(self, child: EasyWidget, index: int):
        child.build()
        return self.layout_.insertWidget(index, child.widget_)

    def hide(self):
        self.widget_.hide()

    def show(self):
        self.widget_.show()

    def removeInnerMargins(self):
        for ch in self.children_:
            if isinstance(ch, EasyWidget):
                ch.removeInnerMargins()
                lay = ch.widget_.layout()
                # layout_ != widget_.layout() for splitter
                if lay:
                    lay.setContentsMargins(0, 0, 0, 0)
                    lay.setSpacing(0)

    @classmethod
    def listChildrenWrapper(
        cls,
        children: t.Union[t.MutableSequence, EasyWidget],
        maybeNewLayout: str = None,
    ):
        if not isinstance(children, EasyWidget):
            children = cls(children)
        if children.layout_ is None and maybeNewLayout is not None:
            children._resetOpts(children.useSplitter, maybeNewLayout)
        return children

    @classmethod
    def buildMainWindow(
        cls,
        children: t.Union[t.MutableSequence, EasyWidget],
        window: QtWidgets.QMainWindow = None,
        layout="V",
        **kwargs,
    ) -> _HasEasyChild:
        if window is None:
            window = QtWidgets.QMainWindow()
        if isinstance(children, t.MutableSequence):
            children = cls(children, layout=layout, **kwargs)

        children.build()
        window.easyChild = children
        window.setCentralWidget(children.widget_)
        children.removeInnerMargins()
        # _HasEasyChild is mostly defined to allow editors to recognize the
        # custom attribute
        return window  # noqa

    @classmethod
    def buildWidget(
        cls, children: t.MutableSequence | EasyWidget, layout="V", **kwargs
    ):
        builder = cls(children, layout=layout, **kwargs)
        builder.build()
        # See noqa note on buildMainWindow
        retWidget: _HasEasyChild = builder.widget_  # noqa
        retWidget.easyChild = builder
        builder.removeInnerMargins()
        return retWidget

    @classmethod
    def fromPgParam(
        cls, param: Parameter = None, layout="H", **opts
    ) -> EasyWidget | tuple[EasyWidget, Parameter]:
        """
        Creates a form-style EasyWidget (name + edit widget) from pyqtgraph parameter
        options or a parameter directly.

        Parameters
        ----------
        param
            Parameter to use, if it already exists. Otherwise, one is created from
            `opts` and returned.
        layout
            EasyWidget layout
        opts
            If `parameter` is unspecified, a parameter is created from these opts instead
            and returned

        Returns
        -------
        Just the EasyWidget if ``parameter`` is provided, otherwise (EasyWidget, Parameter)
        tuple
        """
        returnParam = False
        if param is None:
            param = Parameter.create(**opts)
            returnParam = True
        try:
            item = param.itemClass(param, 0)
            editWidget = item.makeWidget()
            fns.hookupParamWidget(param, editWidget)
        except AttributeError as ex:
            raise ValueError(
                f"Can only create EasyWidgets from parameters that have an itemClass"
                f" and implement makeWidget(). Problem type: {opts['type']}\n"
                f"Exception: {ex}"
            )
        lbl = QtWidgets.QLabel(opts["name"])
        obj = cls([lbl, editWidget], layout)
        obj.build()
        if returnParam:
            return obj, param
        return obj


class _HasEasyChild(QtWidgets.QMainWindow):
    """
    Provided just for type checking purposes
    """

    easyChild: EasyWidget



from pyqtgraph.Qt import QtWidgets

from qtextras import ParameterEditor, fns
from qtextras.examples.parameditor import LAST_RESULT, MyClass


def test_default_use(monkeypatch):
    pe = MyClass()
    with monkeypatch.context() as m:
        m.setattr(QtWidgets.QMessageBox, "information", lambda *args: None)

        for name, func in pe.nameFunctionMap.items():
            lowerName = name.lower()
            if "btn" in lowerName:
                assert pe.rootParameter.child(name).opts["button"]
            if func.parameters:
                paramToChange = next(iter(func.parameters.values()))
                old = LAST_RESULT.value
                if "changed" in lowerName:
                    paramToChange.sigValueChanged.emit(
                        paramToChange, paramToChange.value()
                    )
                    assert LAST_RESULT.value != old
                if "changing" in lowerName:
                    LAST_RESULT.value = old
                    paramToChange.sigValueChanging.emit(
                        paramToChange, paramToChange.value()
                    )
                    assert LAST_RESULT.value != old


def test_state_manager(tmp_path):
    pe = MyClass()
    for value, filter_ in zip(["value", "default"], (fns.valueIsNotDefault, None)):
        cmpState = fns.parameterValues(
            pe.rootParameter, value=value, valueFilter=filter_
        )
        savedState = (
            pe.getParameterDefaults() if value == "default" else pe.getParameterValues()
        )
        assert savedState == cmpState

    pe = MyClass(directory=tmp_path)
    pe.saveParameterValues(pe.stateManager.getDefaultStateName(), addDefaults=True)
    name, state = pe.stateManager.getDefaultStateName(returnStateDict=True)
    assert name.parent == tmp_path

    dummy = pe.rootParameter.addChild(dict(name="dummy", type="str", value="dummy"))
    dummy.setValue("test")
    assert "dummy" in str(pe.saveParameterValues(blockWrite=True))


def test_shortcut():
    pe = ParameterEditor()
    lastValue = None

    def on_shortcut():
        nonlocal lastValue
        lastValue = 17

    pe.registerFunction(on_shortcut, runActionTemplate={"shortcut": "Ctrl+C"})
    assert pe.nameShortcutMap["on_shortcut"].key().toString() == "Ctrl+C"
    pe.nameShortcutMap["on_shortcut"].activated.emit()
    assert lastValue == 17



import logging as _logging

from qtextras._funcparse import (
    ParameterlessInteractor,
    QtExtrasInteractor,
    bindInteractorOptions,
    FROM_PREV_IO,
)
from qtextras.constants import PrjEnums
from qtextras.misc import *
from qtextras.params import *
from qtextras.widgets import *
from qtextras.fns import *

_logging.addLevelName(PrjEnums.LOG_LVL_ATTN, "ATTN")

__version__ = "0.6.5"



import typing as t


class _NAME_STR(str):
    """
    Helper class just used to distinguish between an object reference to a string
    and a string name of a dict key
    """


T = t.TypeVar("T")


class CompositionMixin:
    """
    Python has great built-in support for subclassing. However, another common use case
    is composition-based inheritence, which involves a has-a relationship rather than
    an is-a relationship. Curently, this involves for-looping over the attributes from
    a compositional object to assign them to self, and even this may fail if these
    attributes aren't callable, i.e. if the reference can go stale. This is a helper
    mixin which allows compositional code to be easily written without for-looping over
    attributes or providing stale references.
    """

    _exposedObjs = []

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        obj._exposedObjs = []
        obj._allowRecurse = True
        return obj

    def exposes(self, obj: T, name: str = None) -> T:
        """
        Adds ``obj`` attributes/methods to ``self`` mro / lookup list for easily making
        ``obj`` transparent to consumers of ``self``

        Parameters
        ----------
        obj
            Object to expose
        name
            Often, object references can be reassigned. In these cases, providing a
            name as well will trigger a lookup of this object from self when needed
            rather than relying on the object reference directly
        """
        # Keep only the name in case reference object is altered
        name = obj if name is None else _NAME_STR(name)
        self._exposedObjs.append(name)
        return obj

    def __getattr__(self, item):
        if not self._allowRecurse:
            # Already came here from a previous call where no attribute was found
            raise AttributeError(item)
        self._allowRecurse = False
        for objOrName in self._exposedObjs:
            if isinstance(objOrName, _NAME_STR):
                objOrName = getattr(self, objOrName)
            if hasattr(objOrName, item):
                self._allowRecurse = True
                return getattr(objOrName, item)
        try:
            # Handled by try-except
            # noinspection PyUnresolvedReferences
            ret = super().__getattr__(item)
        except AttributeError:
            raise AttributeError(item)
        finally:
            self._allowRecurse = True
        return ret



import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtWidgets

# Taken directly from https://stackoverflow.com/a/20610786/9463643
try:
    from qtconsole.inprocess import QtInProcessKernelManager
    from qtconsole.rich_jupyter_widget import RichJupyterWidget
except Exception:  # noqa
    # Many things can cause this statement to fail: Kernel already spawned elsewhere,
    # IPython not installed, etc. In any case, pyqtgraph console is an OK fallback
    from pyqtgraph.console import ConsoleWidget
else:

    class ConsoleWidget(RichJupyterWidget):
        """
        Convenience class for a live IPython console widget. We can replace the
        standard banner using the customBanner argument
        """

        def __init__(self, text=None, *args, **kwargs):
            if text:
                self.banner = text
            super().__init__(*args, **kwargs)
            self.kernel_manager = kernel_manager = QtInProcessKernelManager()
            kernel_manager.start_kernel()
            # kernel_manager.kernel.gui = 'qt5'
            self.kernel_client = kernel_client = self._kernel_manager.client()
            kernel_client.start_channels()

            def stop():
                kernel_client.stop_channels()
                kernel_manager.shutdown_kernel()

            self.exit_requested.connect(stop)

            namespace = kwargs.get("namespace", {})
            namespace.setdefault("__console__", self)
            self.pushVariables(namespace)
            parent = kwargs.get("parent", None)
            if parent is not None:
                self.setParent(parent)

        def pushVariables(self, variableDict):
            """
            Given a dictionary containing name / value pairs, push those variables to
            the IPython console widget
            """
            self.kernel_manager.kernel.shell.push(variableDict)

        def clearTerminal(self):
            """
            Clears the terminal
            """
            self._control.clear()

        def printText(self, text):
            """
            Prints some plain text to the console
            """
            self._append_plain_text(text)

        def executeCommand(self, command):
            """
            Execute a command in the frame of the console widget
            """
            self._execute(command, False)


# Sometimes, exceptions in the Ipython stack make it difficult to spawn an ipython
# console even if it is importable Wrap around this issue with a Pyqtgraph console if
# needed
def safeSpawnDevConsole(window: QtWidgets.QMainWindow = None, **consoleLocals):
    """
    Opens a console that allows dynamic interaction with current variables. If IPython
    is on your system, a qt console will be loaded. Otherwise, a (less capable)
    standard pyqtgraph console will be used.
    """
    # "dict" default is to use repr instead of string for internal elements,
    # so expanding into string here ensures repr is not used
    consoleLocals.update(window=window)
    nsPrintout = [f"{k}: {v}" for k, v in consoleLocals.items()]
    text = f"Starting console with variables:\n" f"{nsPrintout}"
    # Broad exception is fine, fallback is good enough. Too many edge cases to properly
    # diagnose when Pycharm's event loop is sync-able with the Jupyter dev console
    try:
        console = ConsoleWidget(parent=window, namespace=consoleLocals, text=text)
    except Exception:  # noqa
        # Ipy kernel can have issues for many reasons. Always be ready to
        # fall back to traditional console
        console = pg.console.ConsoleWidget(
            parent=window, namespace=consoleLocals, text=text
        )
    console.setWindowFlag(QtCore.Qt.WindowType.Window)
    console.show()
